from __future__ import annotations

from base64 import b64encode
from pathlib import Path
from typing import get_args

from htmltools import TagChild, TagList, tags

from .._docstring import add_example, no_example
from .._utils import private_random_int
from ._busy_spinner_types import BusySpinnerType
from ._card import CardItem

__all__ = (
    "options",
    "use",
)


@add_example(ex_dir="../api-examples/busy_indicators")
def options(
    *,
    spinner_type: BusySpinnerType | Path | None = None,
    spinner_color: str | None = None,
    spinner_size: str | None = None,
    spinner_delay: str | None = None,
    spinner_selector: str | None = None,
    fade_opacity: float | None = None,
    fade_selector: str | None = None,
    pulse_background: str | None = None,
    pulse_height: str | None = None,
    pulse_speed: str | None = None,
) -> CardItem:
    """
    Customize spinning busy indicators.

    Busy indicators provide a visual cue to users when the server is busy calculating
    outputs or otherwise performing tasks (e.g., producing downloads). This function
    allows you to customize the appearance of those busy indicators. To apply the
    customization, include the result of this function inside the app's UI.

    Parameters
    ----------
    spinner_type
        The type of spinner. Pre-bundled types are listed in the `BusySpinnerType`
        type.

        A `Path` to a local SVG file can also be provided. The SVG should adhere
        to the following rules:
        * The SVG itself should contain the animation.
        * It should avoid absolute sizes (the spinner's containing DOM element size is
            set in CSS by `spinner_size`, so it should fill that container).
        * It should avoid setting absolute colors (the spinner's containing DOM
            element color is set in CSS by `spinner_color`, so it should inherit that
            color).
    spinner_color
        The color of the spinner. This can be any valid CSS color. Defaults to the
        app's "primary" color (if Bootstrap is on the page).
    spinner_size
        The size of the spinner. This can be any valid CSS size.
    spinner_delay
        The amount of time to wait before showing the spinner. This can be any valid
        CSS time and can useful for not showing the spinner if the computation
        finishes quickly.
    spinner_selector
        A character string containing a CSS selector for scoping the spinner
        customization. The default (`None`) will apply the spinner customization to the
        parent element of the spinner.
    fade_opacity
        The opacity (a number between 0 and 1) for recalculating output. Set to 1 to
        "disable" the fade.
    fade_selector
        A string containing a CSS selector for scoping the fade customization. The
        default (`None`) applies the fade customization to the parent element.
    pulse_background
        A CCS background definition for the pulse. The default uses a
        [linear-gradient](https://developer.mozilla.org/en-US/docs/Web/CSS/gradient/linear-gradient)
        of the theme's indigo, purple, and pink colors.
    pulse_height
        The height of the pulsing banner. This can be any valid CSS size.
    pulse_speed
        The speed of the pulsing banner. This can be any valid CSS time.

    See Also
    --------
    * :func:`~shiny.ui.busy_indicators.use` for enabling/disabling busy indicators.
    """

    res = TagList(
        spinner_options(
            type=spinner_type,
            color=spinner_color,
            size=spinner_size,
            delay=spinner_delay,
            selector=spinner_selector,
        ),
        fade_options(opacity=fade_opacity, selector=fade_selector),
        pulse_options(
            background=pulse_background,
            height=pulse_height,
            speed=pulse_speed,
        ),
    )

    return CardItem(res)


def spinner_options(
    *,
    type: BusySpinnerType | Path | None = None,
    color: str | None = None,
    size: str | None = None,
    delay: str | None = None,
    selector: str | None = None,
) -> TagChild:
    if (
        type is None
        and color is None
        and size is None
        and delay is None
        and selector is None
    ):
        return None

    url = None
    if type is not None:
        if isinstance(type, Path):
            with open(type, "rb") as f:
                type64 = b64encode(f.read()).decode()
                url = f"url('data:image/svg+xml;base64,{type64}')"
        else:
            if type not in get_args(BusySpinnerType):
                raise ValueError(f"Invalid spinner type: {type}")
            url = f"url('spinners/{type}.svg')"

    css_vars = (
        (f"--shiny-spinner-url: {url};" if url else "")
        + (f"--shiny-spinner-color: {color};" if color else "")
        + (f"--shiny-spinner-size: {size};" if size else "")
        + (f"--shiny-spinner-delay: {delay};" if delay else "")
    )

    id = None
    if selector is None:
        id = f"spinner-options-{private_random_int(1000, 1000000)}"
        selector = f":has(> #{id})"

    return tags.style(f"{selector} {{ {css_vars} }}", id=id)


def pulse_options(
    *,
    background: str | None = None,
    height: str | None = None,
    speed: str | None = None,
) -> TagChild:
    if background is None and height is None and speed is None:
        return None

    css_vars = (
        (f"--shiny-pulse-background: {background};" if background else "")
        + (f"--shiny-pulse-height: {height};" if height else "")
        + (f"--shiny-pulse-speed: {speed};" if speed else "")
    )

    return tags.style(":root {" + css_vars + "}")


def fade_options(
    *,
    opacity: float | None = None,
    selector: str | None = None,
) -> TagChild:
    if opacity is None and selector is None:
        return None

    css_vars = f"--shiny-fade-opacity: {opacity};" if opacity else ""

    id = None
    if selector is None:
        id = f"fade-options-{private_random_int(1000, 1000000)}"
        selector = f":has(> #{id})"

    return tags.style(f"{selector} {{ {css_vars} }}", id=id)


@no_example()
def use(*, spinners: bool = True, pulse: bool = True, fade: bool = True) -> TagList:
    """
    Enable/disable busy indication

    Busy indicators provide a visual cue to users when the server is busy calculating
    outputs or otherwise performing tasks (e.g., producing downloads). When enabled
    (they are by default), a spinner is shown on each calculating/recalculating output,
    and a pulsing banner is shown at the top of the page when the app is otherwise busy.
    To disable, include the result of this function in anywhere in the app's UI.

    Parameters
    ----------
    spinners
        Whether to show a spinner on each calculating/recalculating output.
    pulse
        Whether to show a pulsing banner at the top of the page when the app is
        busy.
    fade
        Whether to fade recalculating outputs. A value of `False` is equivalent to
        `shiny.ui.busy_indicators.options(fade_opacity=1)`.

    Note
    ----
    When both `spinners` and `pulse` are set to `True`, the pulse is disabled when
    spinner(s) are active.
    When both `spinners` and `pulse` are set to `False`, no busy indication is shown
    (other than the gray-ing out of recalculating outputs).

    See Also
    --------
    * :func:`~shiny.ui.busy_indicators.options` for customizing busy indicators.
    """

    # TODO: it'd be nice if htmltools had something like a page_attrs() that allowed us
    # to do this without needing to inject JS into the head.
    attrs = {"shinyBusySpinners": spinners, "shinyBusyPulse": pulse}
    js = ""
    for key, value in attrs.items():
        if value:
            js += f"document.documentElement.dataset.{key} = true;"
        else:
            js += f"delete document.documentElement.dataset.{key};"

    return TagList(
        tags.script(js),
        None if fade else fade_options(opacity=1),
    )



from __future__ import annotations

import os

from htmltools import HTMLDependency

from . import __version__
from .ui._html_deps_py_shiny import busy_indicators_dep


def shiny_deps(include_css: bool = True) -> list[HTMLDependency]:
    deps = [
        HTMLDependency(
            name="shiny",
            version=__version__,
            source={"package": "shiny", "subdir": "www/shared/"},
            script={"src": "shiny.js"},
            # This CSS is now rendered against default Bootstrap
            stylesheet={"href": "shiny.min.css"} if include_css else None,
        ),
        busy_indicators_dep(),
    ]

    if os.getenv("SHINY_DEV_MODE") == "1":
        deps.append(
            HTMLDependency(
                "shiny-devmode",
                version=__version__,
                head="<script>window.__SHINY_DEV_MODE__ = true;</script>",
            )
        )

    return deps


def jquery_deps() -> HTMLDependency:
    return HTMLDependency(
        name="jquery",
        version="3.6.0",
        source={"package": "shiny", "subdir": "www/shared/jquery/"},
        script={"src": "jquery-3.6.0.min.js"},
    )


# Shiny doesn't (currently) use requirejs directly, but it does include it because a
# custom requirejs setup is need to get HTMLDependency()s (i.e., loading JS via <script>
# tags) to be usable. At the moment, we're just setting `window.define.amd=false` after
# loading requirejs so that the typical UMD pattern won't result in an anonymous
# define() error.
# https://requirejs.org/docs/errors.html#mismatch
# https://github.com/umdjs/umd
#
# Someday, we may want the same/similar thing in R, but this definitely more of an
# immediate issue for Python since many Jupyter extensions use requirejs.
def require_deps() -> HTMLDependency:
    return HTMLDependency(
        name="requirejs",
        version="2.3.6",
        source={"package": "shiny", "subdir": "www/shared/requirejs/"},
        script={"src": "require.min.js"},
    )



from __future__ import annotations

from astropy.table import Table  # pyright: ignore[reportMissingTypeStubs]

from shiny.express import render, ui

t = Table()
t["a"] = [1, 2, 3, 4]
t["b"] = ["a", "b", "c", "d"]


ui.h2("Original Data")


@render.code
def code_original():
    return str(type(t))


@render.data_frame  # pyright: ignore[reportArgumentType]
def df_astropy():
    return t


@render.code
def code_astropy():
    return str(
        type(
            df_astropy.data_view()  # pyright: ignore[reportUnknownArgumentType,reportUnknownMemberType]
        )
    )


ui.h2(ui.code(".data()"))


@render.data_frame
def df_data():  # pyright: ignore[reportUnknownParameterType]
    return df_astropy.data_view()  # pyright: ignore[reportUnknownVariableType]


@render.code
def code_data():
    return str(
        type(
            df_data.data_view()  # pyright: ignore[reportUnknownArgumentType,reportUnknownMemberType]
        )
    )



import functools
import json

import plotly.graph_objects as go
from shinywidgets import render_widget

from shiny import reactive


# Return a hash of an arbitrary object, including nested dicts, lists, and numpy/pandas
# data structures. Uses json.dumps() internally.
def _hash_anything(obj):
    return hash(json.dumps(obj, sort_keys=True, default=_to_json_repr))


def _to_json_repr(obj):
    # If has to_json(), use that; make sure it's callable
    if hasattr(obj, "to_json") and callable(obj.to_json):
        return json.loads(obj.to_json())
    if hasattr(obj, "to_list") and callable(obj.to_list):
        return obj.to_list()
    if hasattr(obj, "tolist") and callable(obj.tolist):
        return obj.tolist()
    if hasattr(obj, "to_dict") and callable(obj.to_dict):
        return obj.to_dict()
    raise TypeError(f"Object of type {type(obj)} is not JSON serializable")


def render_plotly_streaming(
    fn=None, *, recreate_key=lambda: None, update=("layout", "data")
):
    """Custom decorator for Plotly streaming plots. This is similar to
    shinywidgets.render_widget, except:

    1. You return simply a Figure, not FigureWidget.
    2. On reactive invalidation, the figure is updated in-place, rather than recreated
       from scratch.

    Parameters
    ----------
    recreate_key : callable, optional
        A function that returns a hashable object. If the value returned by this
        function changes, the plot will be recreated from scratch. This is useful for
        changes that render_plotly_streaming can't handle well, such as changing the
        number of traces in a plot.
    """

    if fn is not None:
        return render_plotly_streaming(recreate_key=recreate_key)(fn)

    def decorator(func):
        @deduplicate
        def recreate_trigger():
            return _hash_anything(recreate_key())

        @render_widget
        @functools.wraps(func)
        def wrapper():
            recreate_trigger()

            with reactive.isolate():
                fig = func()
                widget = go.FigureWidget(fig)

            @reactive.effect
            def update_plotly_data():
                f_new = func()
                with widget.batch_update():
                    if "layout" in update:
                        widget.update_layout(f_new.layout)
                    if "data" in update:
                        for old, new in zip(widget.data, f_new.data):
                            old.update(new)

            reactive.get_current_context().on_invalidate(update_plotly_data.destroy)

            return widget

        return wrapper

    return decorator


def deduplicate(func):
    with reactive.isolate():
        rv = reactive.value(func())

    @reactive.effect
    def update():
        x = func()
        with reactive.isolate():
            if x != rv():
                rv.set(x)

    @reactive.calc
    @functools.wraps(func)
    def wrapper():
        return rv()

    return wrapper



from pathlib import Path

from shiny import App, Inputs, Outputs, Session, ui

app_ui = ui.page_fluid(
    ui.tags.link(href="css/styles.css", rel="stylesheet"),
    ui.tags.div(
        "If you see this text, it failed",
        id="target",
        style="background-color: red;",
    ),
    ui.tags.script(src="js/changetext.js"),
    ui.tags.div(
        "This box should be green: ",
        ui.tags.div(
            id="box",
            style="width: 100px; height:100px; border: 1px solid black;",
        ),
    ),
    "There should be a slider below: ",
    ui.input_slider("n", "N", min=1, max=100, value=50),
)


def server(input: Inputs, output: Outputs, session: Session):
    pass


app_dir = Path(__file__).parent.resolve()
app = App(app_ui, server, static_assets=str(app_dir / "www"))



import typing

import pandas as pd

from shiny import App, Inputs, Outputs, Session, reactive, render, req, ui
from shiny.types import FileInfo

app_ui = ui.page_fluid(
    ui.input_file("file1", "Choose CSV File", accept=[".csv"], multiple=False),
    ui.input_checkbox_group(
        "stats",
        "Summary Stats",
        choices=["Row Count", "Column Count", "Column Names"],
        selected=["Row Count", "Column Count", "Column Names"],
    ),
    ui.output_table("summary"),
    ui.input_file("file2", "Multiple files", multiple=True),
    ui.output_text_verbatim("file2_info", placeholder=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.calc
    def parsed_file():
        file: typing.Union[typing.List["FileInfo"], None] = input.file1()
        if file is None:
            return pd.DataFrame()
        return pd.read_csv(  # pyright: ignore[reportUnknownMemberType]
            file[0]["datapath"]
        )

    @render.table
    def summary():
        df = parsed_file()

        if df.empty:
            return pd.DataFrame()

        # Get the row count, column count, and column names of the DataFrame
        row_count = df.shape[0]
        column_count = df.shape[1]
        names: list[str] = (
            df.columns.tolist()  # pyright: ignore[reportUnknownMemberType]
        )
        column_names = ", ".join(str(name) for name in names)

        # Create a new DataFrame to display the information
        info_df = pd.DataFrame(
            {
                "Row Count": [row_count],
                "Column Count": [column_count],
                "Column Names": [column_names],
            }
        )

        # input.stats() is a list of strings; subset the columns based on the selected
        # checkboxes
        return info_df.loc[:, input.stats()]

    @render.text
    def file2_info():
        file2: typing.Union[typing.List["FileInfo"], None] = input.file2()
        if not file2:
            req(file2)
            return

        req(file2)

        ret = [
            f"File name: {file['name']}\nFile type: {file['type']}\nFile size: {file['size']} bytes"
            for file in file2
        ]

        return "\n---\n".join(ret)


app = App(app_ui, server)



# pyright: reportUnknownMemberType=false

import re

from conftest import create_example_fixture
from playwright.sync_api import Page, expect

from shiny.run import ShinyAppProc

cpuinfo_app = create_example_fixture("cpuinfo")


def test_cpuinfo(page: Page, cpuinfo_app: ShinyAppProc):
    page.goto(cpuinfo_app.url)
    plot = page.locator("#plot")
    expect(plot).to_have_class(re.compile(r"\bshiny-bound-output\b"))



from shiny import App, Inputs, Outputs, Session, reactive, render, ui

items = [
    ui.accordion_panel(f"Section {letter}", f"Some narrative for section {letter}")
    for letter in "ABCDE"
]

app_ui = ui.page_fluid(
    # Provide an id to create a shiny input binding
    ui.accordion(*items, id="acc"),
    ui.h4("Accordion:"),
    ui.output_text_verbatim("acc_val", placeholder=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    def _():
        print(input.acc())

    @render.text
    def acc_val():
        return "input.acc(): " + str(input.acc())


app = App(app_ui, server)



from shiny.express import ui


def navset_sidebar():
    from shiny import ui as core_ui

    return core_ui.sidebar(core_ui.markdown("Sidebar content"))


navset_bar_infos = [
    (
        "fixed-top",
        {
            "title": "navset_bar_fixed_top_position_selected",
            "id": "navset_bar_fixed_top_position_selected",
            "position": "fixed-top",
            "selected": "B",
        },
    ),
    (
        "fixed-bottom",
        {
            "title": "navset_bar_header_footer_fixed_bottom_position",
            "id": "navset_bar_header_footer_fixed_bottom_position",
            "header": "Header",
            "footer": "Footer",
            "position": "fixed-bottom",
        },
    ),
    (
        "sticky-top",
        {
            "title": "navset_bar_with_sidebar_collapsible_bg_inverse",
            "id": "navset_bar_with_sidebar_collapsible_bg_inverse",
            "sidebar": navset_sidebar(),
            "collapsible": True,
            "bg": "DodgerBlue",
            "inverse": True,
            "position": "sticky-top",
        },
    ),
    (
        "fixed",
        {
            "title": "navset_bar_collapsible_underline_fixed_gap",
            "id": "navset_bar_collapsible_underline_fixed_gap",
            "collapsible": False,
            "underline": True,
            "fluid": False,
            "gap": "50px",
        },
    ),
]

# Add extra spaces so that the navset_tab is below the fixed-top navset_bar
ui.br()
ui.br()
ui.br()
ui.br()

# TODO-karan; Put each navset_bar into a navpanel within a navset_tab (similar to the navsets_kitchensink app)
with ui.navset_tab(id="navsets_collection"):
    for tab_name, navset_args in navset_bar_infos:
        with ui.nav_panel(tab_name):

            with ui.card(style="position: relative;"):

                with ui.navset_bar(
                    **navset_args  # pyright: ignore[reportArgumentType]
                ):
                    with ui.nav_panel("A"):
                        "Panel A content"

                    with ui.nav_panel("B"):
                        "Panel B content"



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_input_text_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    default = controller.InputText(page, "default")
    default.expect_label("Default text input")
    default.expect_value("")
    controller.OutputCode(page, "default_txt").expect.to_have_text("")

    placeholder = controller.InputText(page, "placeholder")
    placeholder.expect_placeholder("Enter text here")
    placeholder.expect_value("")
    placeholder.expect_autocomplete("off")

    width = controller.InputText(page, "width")
    width.expect_width("200px")
    width.expect_value("Custom width input")

    autocomplete = controller.InputText(page, "autocomplete")
    autocomplete.expect_autocomplete("on")

    spellcheck = controller.InputText(page, "spellcheck")
    spellcheck.expect_spellcheck("true")
    spellcheck.expect_value("paticular")
    controller.OutputCode(page, "spellcheck_txt").expect.to_have_text("paticular")
    spellcheck.set("Some text")
    spellcheck.expect_value("Some text")
    controller.OutputCode(page, "spellcheck_txt").expect.to_have_text("Some text")



from icons import question_circle_fill

from shiny import App, ui

app_ui = ui.page_fluid(
    ui.tooltip(
        ui.input_action_button("btn", "A button", class_="mt-3"),
        "A message",
        id="btn_tooltip",
    ),
    ui.hr(),
    ui.card(
        ui.card_header(
            ui.tooltip(
                ui.span("Card title ", question_circle_fill),
                "Additional info",
                placement="right",
                id="card_tooltip",
            ),
        ),
        "Card body content...",
    ),
)


app = App(app_ui, server=None)



from typing import Union

from shiny.express import render, ui

# Set some Shiny page options
ui.page_opts(title="Hello Chat")

# Create a chat instance, with an initial message
chat = ui.Chat(id="chat")

# Display the chat
chat.ui()


@chat.transform_user_input
async def capitalize(input: str) -> Union[str, None]:
    if input == "return None":
        return None
    elif input == "return custom message":
        await chat.append_message("Custom message")
        return None
    else:
        return input.upper()


@chat.on_user_submit
async def _():
    user = chat.user_input(transform=True)
    await chat.append_message(f"Transformed input: {user}")


"chat.messages():"


@render.code
def message_state():
    return str(chat.messages())


"chat.messages(transform_user='none'):"


@render.code
def message_state2():
    return str(chat.messages(transform_user="none"))



from shiny import reactive
from shiny.express import input, ui

ui.input_action_button("rmv", "Remove UI")
ui.input_text("txt", "Click button above to remove me")


@reactive.effect
@reactive.event(input.rmv)
def _():
    ui.remove_ui(selector="div:has(> #txt)")



import pandas as pd

from shiny import reactive
from shiny.express import input, render, ui
from shiny.types import FileInfo

ui.input_file("file1", "Choose CSV File", accept=[".csv"], multiple=False)
ui.input_checkbox_group(
    "stats",
    "Summary Stats",
    choices=["Row Count", "Column Count", "Column Names"],
    selected=["Row Count", "Column Count", "Column Names"],
)


@reactive.calc
def parsed_file():
    file: list[FileInfo] | None = input.file1()
    if file is None:
        return pd.DataFrame()
    return pd.read_csv(file[0]["datapath"])  # pyright: ignore[reportUnknownMemberType]


@render.table
def summary():
    df = parsed_file()

    if df.empty:
        return pd.DataFrame()

    # Get the row count, column count, and column names of the DataFrame
    row_count = df.shape[0]
    column_count = df.shape[1]
    names = df.columns.tolist()
    column_names = ", ".join(str(name) for name in names)

    # Create a new DataFrame to display the information
    info_df = pd.DataFrame(
        {
            "Row Count": [row_count],
            "Column Count": [column_count],
            "Column Names": [column_names],
        }
    )

    # input.stats() is a list of strings; subset the columns based on the selected
    # checkboxes
    return info_df.loc[:, input.stats()]



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_accordion_kitchensink(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    accordion1 = controller.Accordion(page, "accordion_1")
    accordion1.expect_width("600px")
    accordion1.expect_height("300px")
    accordion1.expect_class("bg-light")
    accordion1.expect_multiple(False)
    accordion1_panel1 = accordion1.accordion_panel("panel1")
    accordion1_panel1.expect_open(True)
    accordion1_panel1.expect_label("Panel 1")
    accordion1_panel1.expect_icon(False)

    accordion1_panel2 = accordion1.accordion_panel("panel2")
    accordion1_panel2.expect_open(False)
    accordion1_panel2.expect_label("Panel 2")
    accordion1_panel2.expect_icon(True)
    accordion1_panel2.set(True)
    accordion1_panel2.expect_open(True)
    accordion1_panel1.expect_open(False)

    accordion2 = controller.Accordion(page, "accordion_2")
    accordion2.expect_width(None)
    accordion2.expect_height(None)
    accordion2.expect_multiple(True)

    accordion2_panel3 = accordion2.accordion_panel("panel3")
    accordion2_panel3.expect_open(True)

    accordion2_panel4 = accordion2.accordion_panel("panel4")
    accordion2_panel4.expect_open(False)
    accordion2_panel4.set(True)
    accordion2_panel4.expect_open(True)
    accordion2_panel3.expect_open(True)



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_switch_kitchensink(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    default = controller.InputSwitch(page, "default")
    default.expect_label("Default switch")
    default.expect_checked(False)

    default_code = controller.OutputCode(page, "default_txt")
    default_code.expect_value("False")

    value = controller.InputSwitch(page, "value")
    value.expect_checked(True)
    controller.OutputCode(page, "value_txt").expect_value("True")

    width = controller.InputSwitch(page, "width")
    width.expect_width("200px")
    controller.OutputCode(page, "width_txt").expect_value("False")
    width.set(True)
    width.expect_checked(True)
    controller.OutputCode(page, "width_txt").expect_value("True")



# pyright:basic
import time

import numpy as np
import seaborn as sns

from shiny import App, module, reactive, render, ui


# -- Reusable card module --
@module.ui
def card_ui(spinner_type, spinner_color, spinner_size):
    return ui.card(
        ui.busy_indicators.options(
            spinner_type=spinner_type,
            spinner_color=spinner_color,
            spinner_size=spinner_size,
        ),
        ui.card_header("Spinner: " + spinner_type),
        ui.output_plot("plot"),
    )


@module.server
def card_server(input, output, session, rerender):
    @render.plot
    def plot():
        rerender()
        time.sleep(0.5)
        sns.lineplot(x=np.arange(100), y=np.random.randn(100))


# -- Main app --
app_ui = ui.page_fillable(
    ui.busy_indicators.options(
        pulse_background="linear-gradient(45deg, blue, red)",
        pulse_height="100px",
        pulse_speed="4s",
    ),
    # ui.busy_indicators.use(spinners=False, pulse=True),
    ui.input_radio_buttons(
        "busy_indicator_type",
        "Choose the indicator type",
        ["spinners", "pulse"],
        inline=True,
    ),
    ui.input_task_button("rerender", "Re-render"),
    ui.output_text_verbatim("counter", placeholder=True),
    ui.layout_columns(
        card_ui("ring", "ring", "red", "10px"),
        card_ui("bars", "bars", "green", "20px"),
        card_ui("dots", "dots", "blue", "30px"),
        card_ui("pulse", "pulse", "olive", "50px"),
        col_widths=6,
    ),
    ui.output_ui("indicator_types_ui"),
)


def server(input, output, session):

    @reactive.calc
    @reactive.event(input.rerender, ignore_none=False)
    def rerender():
        return input.rerender()

    card_server("ring", rerender=rerender)
    card_server("bars", rerender=rerender)
    card_server("dots", rerender=rerender)
    card_server("pulse", rerender=rerender)

    @render.ui
    def indicator_types_ui():
        selected_busy_indicator_type = input.busy_indicator_type()
        return ui.busy_indicators.use(
            spinners=(selected_busy_indicator_type == "spinners"),
            pulse=(selected_busy_indicator_type != "spinners"),
        )

    @render.text
    def counter():
        return str(rerender())


app = App(app_ui, server)



from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.h3("HTTP request headers"),
    ui.output_code("headers", placeholder=True),
    ui.h3("User and groups"),
    ui.output_code("user_groups", placeholder=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.code
    def headers():
        s = ""
        for key, value in session.http_conn.headers.items():
            s += f"{key}: {value}\n"

        return s

    @render.code
    def user_groups():
        return f"session.user: {session.user}\nsession.groups: {session.groups}"


app = App(app_ui, server)



from shiny.express import render


@render.image
def image():
    from pathlib import Path

    dir = Path(__file__).resolve().parent
    img = {"src": str(dir / "posit-logo.png"), "width": "100px"}
    return img



from __future__ import annotations

import asyncio
from typing import (
    Awaitable,
    Callable,
    Generic,
    Literal,
    Optional,
    TypeVar,
    cast,
    overload,
)

from .._docstring import add_example
from .._typing_extensions import ParamSpec
from .._utils import is_async_callable
from .._validation import req
from ._core import Context, flush, lock
from ._reactives import Value, isolate

P = ParamSpec("P")
R = TypeVar("R")


__all__ = (
    "Status",
    "ExtendedTask",
    "extended_task",
)

Status = Literal["initial", "running", "success", "error", "cancelled"]


class DenialContext(Context):
    """
    A context that denies all requests to read reactive sources. We use this to ensure
    that code run outside of reactive.lock doesn't inadvertedly read reactive sources.
    """

    def on_invalidate(self, func: Callable[[], None]) -> None:
        raise RuntimeError(
            "You're not allowed to read reactive sources from inside a Extended Task. "
            "Instead, read the reactive sources before calling the extended task, and "
            "pass them in as function arguments."
        )


class ExtendedTask(Generic[P, R]):
    def __init__(self, func: Callable[P, Awaitable[R]]):
        if not is_async_callable(func):
            raise TypeError("ExtendedTask can only be used with async functions")
        self._func = func
        self._task: Optional[asyncio.Task[R]] = None

        self.status: Value[Status] = Value("initial")
        """
        Reactive value that tracks the current status of the task. The value will be one
        of "initial", "running", "success", "error", or "cancelled".
        """

        self.value: Value[R] = Value()
        """
        Reactive value that tracks the result of the task, if the current status is
        "success". If the status is not "success", the value will be unset, and a silent
        exception will be raised if you try to read it. Calling code should generally
        not read this value directly, but instead use the `result()` method, which is
        designed to behave correctly regardless of the current status.
        """

        self.error: Value[BaseException] = Value()
        """
        Reactive value that tracks the error raised by the task, if the current status
        is "error". If the status is not "error", the value will be unset, and a silent
        exception will be raised if you try to read it. Calling code should generally
        not read this value directly, but instead use the `result()` method, which is
        designed to behave correctly regardless of the current status.
        """

        # If invoked while a previous invocation is still running, we queue up.
        self._invocation_queue: list[Callable[[], None]] = []

    def cancel(self) -> None:
        """
        Cancel the current invocation, if any. If there are pending invocations, cancel
        those too.
        """
        self._invocation_queue.clear()
        if self._task is not None:
            self._task.cancel()

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        self.invoke(*args, **kwargs)

    def invoke(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """
        Request execution of the slow computation. If there's already a computation in
        progress, this will queue up the new invocation to be run after the current one.
        The arguments to this function are passed to the underlying function.

        Returns
        -------
        :
            Immediately returns `None`. The results of the computation will be available
            via the `result()` method.
        """
        with isolate():
            if self.status() == "running" or len(self._invocation_queue) > 0:
                self._invocation_queue.append(lambda: self._invoke(*args, **kwargs))
            else:
                self._invoke(*args, **kwargs)

    def _invoke(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """
        Internal implementation of `invoke()`. The public wrapper protects this method
        from being called while a previous invocation is still running.
        """
        assert self._task is None

        self.status.set("running")
        self.value.unset()
        self.error.unset()
        self._task = asyncio.create_task(self._execution_wrapper(*args, **kwargs))
        self._task.add_done_callback(self._done_callback)

    async def _execution_wrapper(self, *args: P.args, **kwargs: P.kwargs) -> R:
        """
        Wraps the user code in a context that denies access to reactive sources.
        """
        with DenialContext()():
            return await self._func(*args, **kwargs)

    def _done_callback(self, task: asyncio.Task[R]) -> None:
        """
        Several things must happen when a task finishes:
        1. Update the status and value/error.
        2. Reactive flush, so that reactive dependencies can move forward.
        3. If there are queued up invocations, run the next one.
        """

        if self._task is not task:
            return  # This task was cancelled, and a new one was started

        # If the task failed, save the error in self.error.
        # If it succeeded, save the result in self.value.
        self._task = None

        async def _impl():
            async with lock():
                if task.cancelled():
                    self.status.set("cancelled")
                elif task.exception() is not None:
                    self.error.set(cast(BaseException, task.exception()))
                    self.status.set("error")
                else:
                    self.value.set(task.result())
                    self.status.set("success")

                await flush()

                if len(self._invocation_queue) > 0:
                    next_invocation = self._invocation_queue.pop(0)
                    next_invocation()
                    await flush()

        asyncio.create_task(_impl())

    def result(self) -> R:
        """
        Call from a reactive context (e.g. a render function,
        :func:`~shiny.reactive.calc`, or :func:`~shiny.reactive.effect`) to get the
        result of the computation.

        * If the computation has finished successfully, the result will be returned.

        * If the computation has finished with an error, the error will be raised.

        * If the computation has never run, or the most recent run was cancelled, a
          silent exception will be raised that will clear any downstream outputs.

        * If the computation is currently running, a special type of silent exception
          will be raised that will cause the output to visually reflect that calculation
          is in progress.
        """

        if self.status() == "success":
            return self.value.get()
        elif self.status() == "error":
            raise self.error.get()
        elif self.status() == "running":
            req(False, cancel_output="progress")
            # Will never get here, but make type checker happy
            return self.value.get()
        else:  # init, cancelled
            req(False)
            # Will never get here, but make type checker happy
            return self.value.get()


@overload
def extended_task() -> Callable[[Callable[P, Awaitable[R]]], ExtendedTask[P, R]]: ...


@overload
def extended_task(func: Callable[P, Awaitable[R]]) -> ExtendedTask[P, R]: ...


@add_example()
def extended_task(
    func: Optional[Callable[P, Awaitable[R]]] = None
) -> ExtendedTask[P, R] | Callable[[Callable[P, Awaitable[R]]], ExtendedTask[P, R]]:
    """
    Decorator to mark an async function as a slow computation. This will cause the
    function to be run in a background asyncio task, and the results will be available
    via the :class:`~shiny.reactive.ExtendedTask` object returned by the decorator.

    Unlike normal async render functions, effects, and calcs, `extended_task` async
    computations do not block Shiny reactive processing from proceeding. This means that
    they can be used to perform long-running tasks without freezing the session that
    owns them, nor other sessions.

    However, this also means that they cannot access reactive sources. This is because
    processing of inputs and reactivity is not blocked, and so the reactive sources may
    change while the computation is running, which is almost never the desired behavior.
    If any reactive sources are needed by the computation, the decorated function must
    take them as parameters, and the resulting `ExtendedTask` object must be invoked
    with the corresponding arguments.

    Parameters
    ----------
    func
        The function to decorate. It must be ``async``. It can take any parameters and
        return any value (including None).

    Returns
    -------
    :
        An ``ExtendedTask`` object that can be used to check the status of the
        computation and retrieve the result.
    """

    def decorator(func: Callable[P, Awaitable[R]]) -> ExtendedTask[P, R]:
        return ExtendedTask(func)

    if func is not None:
        return decorator(func)
    else:
        return decorator



__all__ = ("input_action_button", "input_action_link")

from typing import Optional

from htmltools import Tag, TagAttrValue, TagChild, css, tags

from .._docstring import add_example
from .._namespaces import resolve_id


@add_example()
@add_example("app-disabled-core.py")
def input_action_button(
    id: str,
    label: TagChild,
    *,
    icon: TagChild = None,
    width: Optional[str] = None,
    disabled: bool = False,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Creates an action button whose value is initially zero, and increments by one each
    time it is pressed.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    icon
        An icon to appear inline with the button/link.
    width
        The CSS width, e.g. '400px', or '100%'
    disabled
        If `True`, the button will not be clickable. Use
        :func:`~shiny.ui.update_action_button` to dynamically enable/disable the button.
    **kwargs
        Attributes to be applied to the button.

    Returns
    -------
    :
        A UI element

    Notes
    ------
    ::: {.callout-note title="Server value"}
    An integer representing the number of clicks.
    :::

    See Also
    --------
    * :func:`~shiny.ui.update_action_button`
    * :func:`~shiny.ui.input_action_link`
    * :func:`~shiny.reactive.event`
    """

    if "_add_ws" not in kwargs:
        kwargs["_add_ws"] = True

    return tags.button(
        {"class": "btn btn-default action-button", "style": css(width=width)},
        icon,
        None if icon is None else " ",
        label,
        id=resolve_id(id),
        type="button",
        disabled="" if disabled else None,
        **kwargs,
    )


@add_example()
def input_action_link(
    id: str,
    label: TagChild,
    *,
    icon: TagChild = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Creates a link whose value is initially zero, and increments by one each time it is
    pressed.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    icon
        An icon to appear inline with the button/link.
    **kwargs
        Attributes to be applied to the link.

    Returns
    -------
    :
        A UI element

    Notes
    ------
    ::: {.callout-note title="Server value"}
    An integer representing the number of clicks.
    :::

    See Also
    --------
    * :func:`~shiny.ui.update_action_link`
    * :func:`~shiny.ui.input_action_button`
    * :func:`~shiny.reactive.event`
    """

    return tags.a(
        {"class": "action-button"},
        icon,
        label,
        id=resolve_id(id),
        href="#",
        **kwargs,
    )



from shiny import reactive
from shiny.express import input, session, ui

ui.input_text("msg", "Enter a message")
ui.input_action_button("submit", "Submit the message")
# It'd be better to use ui.insert_ui() in order to implement this kind of
# functionality...this is just a basic demo of how custom message handling works.
ui.tags.div(id="messages")
ui.tags.script(
    """
    $(function() {
        Shiny.addCustomMessageHandler("append_msg", function(message) {
            $("<p>").text(message.msg).appendTo("#messages");
        });
    });
    """
)


@reactive.effect
@reactive.event(input.submit)
async def _():
    await session.send_custom_message("append_msg", {"msg": input.msg()})



import pytest
from playwright.sync_api import Page, expect
from utils.deploy_utils import (
    local_deploys_app_url_fixture,
    reruns,
    reruns_delay,
    skip_if_not_chrome,
)

app_url = local_deploys_app_url_fixture("shiny_client_console_error")


@skip_if_not_chrome
@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
def test_shiny_client_console_error(page: Page, app_url: str) -> None:
    page.goto(app_url)

    assert page.locator("#same_id").count() == 2
    shiny_error_message = page.locator("shiny-error-message")

    # show the client error message only for local apps
    if "127.0.0.1" in app_url:

        expect(shiny_error_message).not_to_have_count(0)
        expect(shiny_error_message).to_have_attribute(
            "message", 'The following ID was repeated:\n- "same_id": 2 inputs'
        )
        expect(page.get_by_role("button", name="Dismiss all")).to_have_count(1)
        expect(
            page.get_by_role("button", name="Copy error to clipboard")
        ).to_have_count(1)

    # for deployed apps to shinyapps.io or connect hide the client error message
    else:
        expect(shiny_error_message).to_have_count(0)



import pytest
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@pytest.mark.parametrize("tab_name", ["pandas", "polars"])
def test_validate_html_columns(
    page: Page,
    local_app: ShinyAppProc,
    tab_name: str,
) -> None:
    page.goto(local_app.url)

    controller.NavsetCardUnderline(page, "tab").set(tab_name)

    data_frame = controller.OutputDataFrame(page, f"{tab_name}_iris")
    # Data frame with html content in the first two columns; Edit a cell in the third column and try to hit `shift + tab`. It should not submit the edit in the current cell and stay at the current cell (not moving to the second or first column)
    data_frame.expect_cell("1.4", row=0, col=2)
    data_frame.set_cell("152", row=0, col=2, finish_key="Shift+Tab")
    data_frame.expect_cell("1.4", row=0, col=2)
    data_frame.expect_cell_class("cell-edit-editing", row=0, col=2)

    # Data frame with html content in the last two columns; Edit a cell in the third from last column and try to hit `tab`. It should not submit the edit in the current cell and stay at the current cell (not moving to the last two columns)
    data_frame.expect_cell("1.4", row=0, col=2)
    data_frame.set_cell("152", row=0, col=2, finish_key="Tab")
    data_frame.expect_cell("1.4", row=0, col=2)
    data_frame.expect_cell_class("cell-edit-editing", row=0, col=2)



from shiny import App, ui

app_ui = ui.page_fillable(
    ui.panel_title("Hello Shiny Chat"),
    ui.chat_ui("chat"),
    fillable_mobile=True,
)

# Create a welcome message
welcome = ui.markdown(
    """
    Hi! This is a simple Shiny `Chat` UI. Enter a message below and I will
    simply repeat it back to you. For more examples, see this
    [folder of examples](https://github.com/posit-dev/py-shiny/tree/main/examples/chat).
    """
)


def server(input, output, session):
    chat = ui.Chat(id="chat", messages=[welcome])

    # Define a callback to run when the user submits a message
    @chat.on_user_submit
    async def _():
        # Get the user's input
        user = chat.user_input()
        # Append a response to the chat
        await chat.append_message(f"You said: {user}")


app = App(app_ui, server)



import json
from dataclasses import asdict, dataclass
from typing import Literal, TypedDict


@dataclass
class QuartodocObject:
    project: str
    version: str
    count: int
    items: list["QuartodocObjectItem"]


@dataclass
class QuartodocObjectItem:
    name: str
    domain: str

    # function: "shiny.ui.page_sidebar"
    # class: "shiny.render.renderer._renderer.Renderer"
    # attribute: "shiny.render.renderer._renderer.Renderer.output_id"
    role: Literal["function", "class", "attribute", "module"]
    priority: str
    uri: str
    dispname: str


def read_objects_file(path: str) -> QuartodocObject:
    with open(path) as file:
        content = json.load(file)
    items = [QuartodocObjectItem(**item) for item in content.pop("items")]
    return QuartodocObject(**content, items=items)


def write_objects_file(objects: QuartodocObject, path: str) -> None:
    with open(path, "w") as file:
        json.dump(objects, file, indent=4, default=lambda dc: dc.__dict__)


print("\nCombining objects json files...")
objects_core = read_objects_file("_objects_core.json")
objects_express = read_objects_file("_objects_express.json")
objects_test = read_objects_file("_objects_test.json")

items_map: dict[str, QuartodocObjectItem] = {}

for item in [*objects_core.items, *objects_express.items, *objects_test.items]:
    if item.name in items_map:
        continue
    items_map[item.name] = item

objects_ret = QuartodocObject(
    project="shiny",
    version="1",
    count=len(items_map.values()),
    items=[*items_map.values()],
)


print("Core:", objects_core.count)
print("Express:", objects_express.count)
print("Testing:", objects_test.count)
print("Combined:", objects_ret.count)

# Save combined objects file info
write_objects_file(objects_ret, "objects.json")



from shiny.express import input, render, ui

ui.input_text_area(
    "caption_regular",
    "Caption:",
    "Data summary\nwith\nmultiple\nlines",
)


@render.text
def value_regular():
    return input.caption_regular()


ui.input_text_area(
    "caption_autoresize",
    ui.markdown("Caption (w/ `autoresize=True`):"),
    "Data summary\nwith\nmultiple\nlines",
    autoresize=True,
)


@render.text
def value_autoresize():
    return input.caption_autoresize()



# TODO-barret; Test this app!

from __future__ import annotations

import palmerpenguins  # pyright: ignore[reportMissingTypeStubs]
import pandas as pd

from shiny import reactive
from shiny.express import input, render, ui

pd_data = palmerpenguins.load_penguins_raw()


with ui.card():

    ui.input_action_button("shift_btn", "Shift data")
    ui.input_action_button("different_btn", "Change data set")

    ui.h4("Data")

    @render.data_frame
    def df():
        return render.DataGrid(
            pd_data.iloc[:, 0:2],
            selection_mode="rows",
            filters=True,
            editable=True,
        )

    ui.h4("Selected data")

    @render.data_frame
    def df_selected():
        return df.data_view(selected=True)


# Reactive value to store the un-subsetted data
full_data = reactive.value(pd_data)


@reactive.effect
@reactive.event(input.shift_btn)
async def shift_data():

    if not input.shift_btn():
        raise ValueError("update_btn must exist")

    k = 2
    shift = (k * input.shift_btn()) % full_data().shape[0]
    subsetted_data = full_data().iloc[(0 + shift) : (k + shift), 0:2]
    await df.update_data(subsetted_data)
    return


@reactive.effect
@reactive.event(input.different_btn)
async def different_data():

    if not input.different_btn():
        raise ValueError("different_btn must exist")

    print(
        "input.different_btn()", input.different_btn(), input.different_btn() % 2 == 0
    )
    if input.different_btn() % 2 == 0:
        await df.update_data(pd_data)
        full_data.set(pd_data)
        return

    new_df = pd.DataFrame(
        {
            "Letter": [
                "a",
                "b",
                "c",
                "d",
                "e",
                "f",
                "g",
                "h",
                "i",
                "j",
                "k",
                "l",
                "m",
                "n",
                "o",
                "p",
                "q",
                "r",
                "s",
                "t",
                "u",
                "v",
                "w",
                "x",
                "y",
                "z",
            ],
            "Negative index": [-1 * i for i in range(1, 27)],
        },
    )
    await df.update_data(new_df)
    full_data.set(new_df)
    return



import random

from shiny import reactive, ui
from shiny.express import input


def make_panel(letter):
    return ui.accordion_panel(
        f"Section {letter}", f"Some narrative for section {letter}"
    )


ui.input_action_button("add_panel", "Add random panel", class_="mt-3 mb-3")
ui.accordion(*[make_panel(letter) for letter in "ABCDE"], id="acc", multiple=True)


@reactive.effect
@reactive.event(input.add_panel)
def _():
    ui.insert_accordion_panel("acc", make_panel(str(random.randint(0, 10000))))



from dataclasses import dataclass
from datetime import date, datetime, timedelta

import pandas as pd
from htmltools import HTML, strong

from shiny import App, Inputs, render, ui


class C:
    x: int

    def __init__(self, x: int):
        self.x = x

    def __str__(self):
        return f"<{self.__class__.__name__} object>"


@dataclass
class D:
    y: int


DATA = {  # pyright: ignore[reportUnknownVariableType]
    "num": [1, 2],
    "chr": ["a", "b"],
    "cat": pd.Series(["c", "d"], dtype="category"),
    "bool": [True, False],
    "date": [date(2000, 1, 2)] * 2,
    "datetime": [datetime(2000, 1, 2)] * 2,
    "duration": [timedelta(weeks=1), timedelta(days=7)],
    "html": [strong("bolded content")] * 2,
    "html_str": [HTML("<strong>bolded string</strong>")] * 2,
    "struct": [{"x": 1}, {"x": 2}],
    "arr": [[1, 2], [3, 4]],
    "object": [C(1), D(2)],
}

df = pd.DataFrame(DATA)


@render.data_frame
def df_html():
    return df


app_ui = ui.page_fillable(
    ui.card(
        ui.output_data_frame("df_html"),
    )
)


def server(input: Inputs):
    @render.data_frame
    def df_html():
        return df


app = App(app_ui, server)



from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.input_action_button("show", "Show"),
    " ",
    ui.input_action_button("remove", "Remove"),
)


def server(input: Inputs, output: Outputs, session: Session):
    ids: list[str] = []
    n: int = 0

    @reactive.effect
    @reactive.event(input.show)
    def _():
        nonlocal ids
        nonlocal n
        # Save the ID for removal later
        id = ui.notification_show("Message " + str(n), duration=None)
        ids.append(id)
        n += 1

    @reactive.effect
    @reactive.event(input.remove)
    def _():
        nonlocal ids
        if ids:
            ui.notification_remove(ids.pop())


app = App(app_ui, server, debug=True)



from __future__ import annotations

import os
import sys
import textwrap
from pathlib import Path
from typing import Optional

import click
import questionary


def directory_prompt(
    dest_dir: Optional[Path | str | None] = None,
    default_dir: Optional[str | None] = None,
) -> Path:
    if dest_dir is not None:
        dest_dir = Path(dest_dir)

        if dest_dir.exists() and dest_dir.is_file():
            click.echo(
                cli_danger(
                    f"Error: Destination directory {cli_field(str(dest_dir))} is a file, not a directory."
                )
            )
            sys.exit(1)
        return dest_dir

    app_dir = questionary.path(
        "Enter destination directory:",
        default=path_rel_wd(default_dir) if default_dir is not None else "./",
        only_directories=True,
    ).ask()

    if app_dir is None:
        sys.exit(1)

    # Perform not-a-file check on the selected `app_dir`
    return directory_prompt(dest_dir=app_dir)


def path_rel_wd(*path: str):
    """
    Path relative to the working directory, formatted for the current OS
    """
    return os.path.join(".", *(path or [""]))


# CLI Style Helpers -----------------------------------------------------------------


def cli_field(x: str):
    return click.style(x, fg="cyan")


def cli_bold(x: str):
    return click.style(x, bold=True)


def cli_ital(x: str):
    return click.style(x, italic=True)


def cli_input(x: str):
    return click.style(x, fg="green")


def cli_code(x: str):
    return click.style("`" + x + "`", fg="magenta")


def cli_verbatim(x: str | list[str], indent: int = 2):
    lines = [click.style(line, fg="cyan") for line in x if line != ""]
    return textwrap.indent("\n".join(lines), " " * indent)


def cli_url(x: str):
    return click.style(x, fg="blue", underline=True)


def cli_success(x: str):
    return click.style("\u2713", fg="green") + " " + x


def cli_info(x: str):
    return click.style("\u2139", fg="blue") + " " + x


def cli_action(x: str):
    return click.style("→", fg="blue") + " " + x


def cli_warning(x: str):
    return click.style("!", fg="yellow") + " " + x


def cli_danger(x: str):
    return click.style("\u00d7", fg="red") + " " + x


def cli_wait(x: str):
    return click.style("\u2026", fg="yellow") + " " + x



import pytest
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@pytest.mark.parametrize("dt_name", ["pandas", "polars"])
def test_validate_column_labels(
    page: Page, local_app: ShinyAppProc, dt_name: str
) -> None:
    page.goto(local_app.url)

    w_filters = controller.OutputDataFrame(page, f"{dt_name}-w_filters")
    wo_filters = controller.OutputDataFrame(page, f"{dt_name}-wo_filters")

    ex_labels = [
        "Sample Number",
        "Species",
        "Region",
    ]

    w_filters.expect_column_labels(ex_labels)
    wo_filters.expect_column_labels(ex_labels)



import random

from shiny import App, Inputs, Outputs, Session, reactive, ui


def make_panel(letter: str) -> ui.AccordionPanel:
    return ui.accordion_panel(
        f"Section {letter}", f"Some narrative for section {letter}"
    )


items = [make_panel(letter) for letter in "ABCDE"]

choices = ["A", "B", "C", "D", "E"]
random.shuffle(choices)

app_ui = ui.page_fluid(
    ui.input_action_button(
        "remove_panel",
        f"Remove Section {choices[-1]}",
        class_="mt-3 mb-3",
    ),
    " (Sections randomly picked at server start)",
    ui.accordion(*items, id="acc", multiple=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    # Copy the list for user
    user_choices = [choice for choice in choices]

    @reactive.effect
    @reactive.event(input.remove_panel)
    def _():
        if len(user_choices) == 0:
            ui.notification_show("No more panels to remove!")
            return

        # Remove panel
        ui.remove_accordion_panel("acc", f"Section {user_choices.pop()}")

        label = "No more panels to remove!"
        if len(user_choices) > 0:
            label = f"Remove Section {user_choices[-1]}"
        ui.update_action_button("remove_panel", label=label)


app = App(app_ui, server)



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_tooltip(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    tooltip = controller.Tooltip(page, "tooltip_id")
    tooltip.expect_active(False)
    tooltip.set(True)
    tooltip.expect_active(True)
    tooltip.expect_body("A message")
    tooltip.expect_placement("right")
    tooltip.set(False)
    tooltip.expect_active(False)



import pandas as pd
import seaborn as sns
import shinyswatch

from shiny import App, Inputs, Outputs, Session, reactive, render, req, ui


def app_ui(req):
    dark = True if "dark" in req.query_params else None

    return ui.page_fluid(
        ui.head_content(
            ui.tags.meta(name="viewport", content="width=device-width, initial-scale=1")
        ),
        light_dark_switcher(dark),
        ui.input_select("dataset", "Dataset", sns.get_dataset_names()),
        ui.input_select(
            "selection_mode",
            "Selection mode",
            {
                "none": "(None)",
                "row": "Single row",
                "rows": "Multiple rows",
            },
            selected="rows",
        ),
        ui.input_switch("editable", "Edit", False),
        ui.input_switch("filters", "Filters", True),
        ui.input_switch("gridstyle", "Grid", True),
        ui.input_switch("fullwidth", "Take full width", True),
        ui.output_data_frame("grid"),
        ui.panel_fixed(
            ui.output_code("detail"),
            right="10px",
            bottom="10px",
        ),
        theme=shinyswatch.theme.darkly if dark else None,
        class_="p-3",
    )


def light_dark_switcher(dark):
    return (
        ui.div(
            ui.a(
                {"class": "btn-primary" if not dark else "btn-outline-primary"},
                "Light",
                href="?" if dark else None,
                class_="btn",
            ),
            ui.a(
                {"class": "btn-primary" if dark else "btn-outline-primary"},
                "Dark",
                href="?dark=1" if not dark else None,
                class_="btn",
            ),
            class_="float-end btn-group",
        ),
    )


def server(input: Inputs, output: Outputs, session: Session):
    df: reactive.value[pd.DataFrame] = reactive.value()

    @reactive.effect
    def update_df():
        return df.set(sns.load_dataset(req(input.dataset())))

    @render.data_frame
    def grid():
        height = 350
        width = "100%" if input.fullwidth() else "fit-content"
        if input.gridstyle():
            return render.DataGrid(
                df(),
                width=width,
                height=height,
                filters=input.filters(),
                editable=input.editable(),
                selection_mode=input.selection_mode(),
            )
        else:
            return render.DataTable(
                df(),
                width=width,
                height=height,
                filters=input.filters(),
                editable=input.editable(),
                selection_mode=input.selection_mode(),
            )

    @reactive.effect
    @reactive.event(input.grid_cell_edit)
    def handle_edit():
        edit = input.grid_cell_edit()
        df_copy = df().copy()
        df_copy.iat[edit["row"], edit["col"]] = edit["new_value"]
        df.set(df_copy)

    @render.code
    def detail():
        selected_rows = grid.cell_selection()["rows"]
        if len(selected_rows) > 0:
            return df().iloc[list(selected_rows)]


app = App(app_ui, server)



from html import escape  # noqa: F401

from shiny import App, Inputs, Outputs, Session, render, ui

states = {
    "East Coast": {"NY": "New York", "NJ": "New Jersey", "CT": "Connecticut"},
    "West Coast": {"WA": "Washington", "OR": "Oregon", "CA": "California"},
    "Midwest": {"MN": "Minnesota", "WI": "Wisconsin", "IA": "Iowa"},
}

app_ui = ui.page_fluid(
    ui.input_selectize(
        "state",
        "Choose a state:",
        states,
        multiple=True,
    ),
    ui.output_text("value"),
    ui.input_selectize(
        "state2",
        "Selectize Options",
        states,
        multiple=True,
        options=(
            {
                "placeholder": "Enter text",
                "render": ui.js_eval(
                    '{option: function(item, escape) {return "<div><strong>Select " + escape(item.label) + "</strong></div>";}}'
                ),
                "create": True,
            }
        ),
    ),
    ui.input_selectize(
        "state3",
        "Selectize plugins",
        states,
        multiple=True,
        options={"plugins": ["clear_button"]},
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.text
    def value():
        return "You choose: " + str(input.state())


app = App(app_ui, server)



import asyncio
from datetime import datetime

from shiny import reactive, render
from shiny.express import input, ui

ui.h5("Current time")


@render.text
def current_time():
    reactive.invalidate_later(1)
    return datetime.now().strftime("%H:%M:%S")


with ui.p():
    "Notice that the time above updates every second, even if you click the button below."


@ui.bind_task_button(button_id="btn")
@reactive.extended_task
async def slow_compute(a: int, b: int) -> int:
    await asyncio.sleep(3)
    return a + b


with ui.layout_sidebar():
    with ui.sidebar():
        ui.input_numeric("x", "x", 1)
        ui.input_numeric("y", "y", 2)
        ui.input_task_button("btn", "Compute, slowly")
        ui.input_action_button("btn_cancel", "Cancel")

    @reactive.effect
    @reactive.event(input.btn, ignore_none=False)
    def handle_click():
        # slow_compute.cancel()
        slow_compute(input.x(), input.y())

    @reactive.effect
    @reactive.event(input.btn_cancel)
    def handle_cancel():
        slow_compute.cancel()

    ui.h5("Sum of x and y")

    @render.text
    def show_result():
        return str(slow_compute.result())



from shiny.express import ui

with ui.hold() as a_card:
    with ui.card():
        "A simple card"

# Always has 2 columns (on non-mobile)
with ui.layout_column_wrap(width=1 / 2):
    a_card
    a_card
    a_card

ui.hr()

# Has three columns when viewport is wider than 750px
with ui.layout_column_wrap(width="250px"):
    a_card
    a_card
    a_card



from shiny import App, render, ui

app_ui = ui.page_fluid(
    ui.input_text("name", "Name", "Socrates"),
    ui.input_text("years", "Years", "470-399 BC"),
    ui.output_ui("person"),
)


def server(input, output, session):

    @render.express
    def person():
        from shiny.express import ui

        with ui.card(class_="mt-3"):
            ui.h3(input.name())
            input.years()


app = App(app_ui, server)



import sys
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Optional, cast

from htmltools import HTML

from ._chat_types import ChatMessage

if TYPE_CHECKING:
    from anthropic.types import Message as AnthropicMessage
    from anthropic.types import MessageStreamEvent

    if sys.version_info >= (3, 9):
        from google.generativeai.types.generation_types import (  # pyright: ignore[reportMissingTypeStubs]
            GenerateContentResponse,
        )
    else:

        class GenerateContentResponse:
            text: str

    from langchain_core.messages import BaseMessage, BaseMessageChunk
    from openai.types.chat import ChatCompletion, ChatCompletionChunk


class BaseMessageNormalizer(ABC):
    @abstractmethod
    def normalize(self, message: Any) -> ChatMessage:
        pass

    @abstractmethod
    def normalize_chunk(self, chunk: Any) -> ChatMessage:
        pass

    @abstractmethod
    def can_normalize(self, message: Any) -> bool:
        pass

    @abstractmethod
    def can_normalize_chunk(self, chunk: Any) -> bool:
        pass


class StringNormalizer(BaseMessageNormalizer):
    def normalize(self, message: Any) -> ChatMessage:
        x = cast(Optional[str], message)
        return ChatMessage(content=x or "", role="assistant")

    def normalize_chunk(self, chunk: Any) -> ChatMessage:
        x = cast(Optional[str], chunk)
        return ChatMessage(content=x or "", role="assistant")

    def can_normalize(self, message: Any) -> bool:
        return isinstance(message, (str, HTML)) or message is None

    def can_normalize_chunk(self, chunk: Any) -> bool:
        return isinstance(chunk, (str, HTML)) or chunk is None


class DictNormalizer(BaseMessageNormalizer):
    def normalize(self, message: Any) -> ChatMessage:
        x = cast("dict[str, Any]", message)
        if "content" not in x:
            raise ValueError("Message must have 'content' key")
        return ChatMessage(content=x["content"], role=x.get("role", "assistant"))

    def normalize_chunk(self, chunk: Any) -> ChatMessage:
        x = cast("dict[str, Any]", chunk)
        if "content" not in x:
            raise ValueError("Message must have 'content' key")
        return ChatMessage(content=x["content"], role=x.get("role", "assistant"))

    def can_normalize(self, message: Any) -> bool:
        return isinstance(message, dict)

    def can_normalize_chunk(self, chunk: Any) -> bool:
        return isinstance(chunk, dict)


class LangChainNormalizer(BaseMessageNormalizer):
    def normalize(self, message: Any) -> ChatMessage:
        x = cast("BaseMessage", message)
        if isinstance(x.content, list):  # type: ignore
            raise ValueError(
                "The `message.content` provided seems to represent numerous messages. "
                "Consider iterating over `message.content` and calling .append_message() on each iteration."
            )
        return ChatMessage(content=x.content, role="assistant")

    def normalize_chunk(self, chunk: Any) -> ChatMessage:
        x = cast("BaseMessageChunk", chunk)
        if isinstance(x.content, list):  # type: ignore
            raise ValueError(
                "The `message.content` provided seems to represent numerous messages. "
                "Consider iterating over `message.content` and calling .append_message() on each iteration."
            )
        return ChatMessage(content=x.content, role="assistant")

    def can_normalize(self, message: Any) -> bool:
        try:
            from langchain_core.messages import BaseMessage

            return isinstance(message, BaseMessage)
        except Exception:
            return False

    def can_normalize_chunk(self, chunk: Any) -> bool:
        try:
            from langchain_core.messages import BaseMessageChunk

            return isinstance(chunk, BaseMessageChunk)
        except Exception:
            return False


class OpenAINormalizer(StringNormalizer):
    def normalize(self, message: Any) -> ChatMessage:
        x = cast("ChatCompletion", message)
        return super().normalize(x.choices[0].message.content)

    def normalize_chunk(self, chunk: Any) -> ChatMessage:
        x = cast("ChatCompletionChunk", chunk)
        return super().normalize_chunk(x.choices[0].delta.content)

    def can_normalize(self, message: Any) -> bool:
        try:
            from openai.types.chat import ChatCompletion

            return isinstance(message, ChatCompletion)
        except Exception:
            return False

    def can_normalize_chunk(self, chunk: Any) -> bool:
        try:
            from openai.types.chat import ChatCompletionChunk

            return isinstance(chunk, ChatCompletionChunk)
        except Exception:
            return False


class AnthropicNormalizer(BaseMessageNormalizer):
    def normalize(self, message: Any) -> ChatMessage:
        x = cast("AnthropicMessage", message)
        content = x.content[0]
        if content.type != "text":
            raise ValueError(
                f"Anthropic message type {content.type} not supported. "
                "Only 'text' type is currently supported"
            )
        return ChatMessage(content=content.text, role="assistant")

    def normalize_chunk(self, chunk: Any) -> ChatMessage:
        x = cast("MessageStreamEvent", chunk)
        content = ""
        if x.type == "content_block_delta":
            if x.delta.type != "text_delta":
                raise ValueError(
                    f"Anthropic message delta type {x.delta.type} not supported. "
                    "Only 'text_delta' type is supported"
                )
            content = x.delta.text

        return ChatMessage(content=content, role="assistant")

    def can_normalize(self, message: Any) -> bool:
        try:
            from anthropic.types import Message as AnthropicMessage

            return isinstance(message, AnthropicMessage)
        except Exception:
            return False

    def can_normalize_chunk(self, chunk: Any) -> bool:
        try:
            from anthropic.types import (
                RawContentBlockDeltaEvent,
                RawContentBlockStartEvent,
                RawContentBlockStopEvent,
                RawMessageDeltaEvent,
                RawMessageStartEvent,
                RawMessageStopEvent,
            )

            # The actual MessageStreamEvent is a generic, so isinstance() can't
            # be used to check the type. Instead, we manually construct the relevant
            # union of relevant classes...
            return (
                isinstance(chunk, RawContentBlockDeltaEvent)
                or isinstance(chunk, RawContentBlockStartEvent)
                or isinstance(chunk, RawContentBlockStopEvent)
                or isinstance(chunk, RawMessageDeltaEvent)
                or isinstance(chunk, RawMessageStartEvent)
                or isinstance(chunk, RawMessageStopEvent)
            )
        except Exception:
            return False


class GoogleNormalizer(BaseMessageNormalizer):
    def normalize(self, message: Any) -> ChatMessage:
        x = cast("GenerateContentResponse", message)
        return ChatMessage(content=x.text, role="assistant")

    def normalize_chunk(self, chunk: Any) -> ChatMessage:
        x = cast("GenerateContentResponse", chunk)
        return ChatMessage(content=x.text, role="assistant")

    def can_normalize(self, message: Any) -> bool:
        try:
            import google.generativeai.types.generation_types as gtypes  # pyright: ignore[reportMissingTypeStubs, reportMissingImports]

            return isinstance(
                message,
                gtypes.GenerateContentResponse,  # pyright: ignore[reportUnknownMemberType]
            )
        except Exception:
            return False

    def can_normalize_chunk(self, chunk: Any) -> bool:
        return self.can_normalize(chunk)


class OllamaNormalizer(DictNormalizer):
    def normalize(self, message: Any) -> ChatMessage:
        x = cast("dict[str, Any]", message["message"])
        return super().normalize(x)

    def normalize_chunk(self, chunk: "dict[str, Any]") -> ChatMessage:
        msg = cast("dict[str, Any]", chunk["message"])
        return super().normalize_chunk(msg)

    def can_normalize(self, message: Any) -> bool:
        if not isinstance(message, dict):
            return False
        if "message" not in message:
            return False
        return super().can_normalize(message["message"])

    def can_normalize_chunk(self, chunk: Any) -> bool:
        return self.can_normalize(chunk)


class NormalizerRegistry:
    def __init__(self) -> None:
        # Order of strategies matters (the 1st one that can normalize the message is used)
        # So make sure to put the most specific strategies first
        self._strategies: dict[str, BaseMessageNormalizer] = {
            "openai": OpenAINormalizer(),
            "anthropic": AnthropicNormalizer(),
            "google": GoogleNormalizer(),
            "langchain": LangChainNormalizer(),
            "ollama": OllamaNormalizer(),
            "dict": DictNormalizer(),
            "string": StringNormalizer(),
        }

    def register(
        self, provider: str, strategy: BaseMessageNormalizer, force: bool = False
    ) -> None:
        if provider in self._strategies:
            if force:
                del self._strategies[provider]
            else:
                raise ValueError(f"Provider {provider} already exists in registry")
        # Update the strategies dict such that the new strategy is the first to be considered
        self._strategies = {provider: strategy, **self._strategies}


message_normalizer_registry = NormalizerRegistry()


def normalize_message(message: Any) -> ChatMessage:
    strategies = message_normalizer_registry._strategies
    for strategy in strategies.values():
        if strategy.can_normalize(message):
            return strategy.normalize(message)
    raise ValueError(
        f"Could not find a normalizer for message of type {type(message)}: {message}. "
        "Consider registering a custom normalizer via shiny.ui._chat_types.registry.register()"
    )


def normalize_message_chunk(chunk: Any) -> ChatMessage:
    strategies = message_normalizer_registry._strategies
    for strategy in strategies.values():
        if strategy.can_normalize_chunk(chunk):
            return strategy.normalize_chunk(chunk)
    raise ValueError(
        f"Could not find a normalizer for message chunk of type {type(chunk)}: {chunk}. "
        "Consider registering a custom normalizer via shiny.ui._chat_types.registry.register()"
    )



import matplotlib.pyplot as plt
import numpy as np

from shiny import App, Inputs, Outputs, Session, reactive, render, ui

app_ui = ui.page_fluid(
    ui.input_slider("n", "Number of observations", min=0, max=1000, value=500),
    ui.input_action_link("go", "Go!"),
    ui.output_plot("plot"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.plot(alt="A histogram")
    # reactive.event() to invalidate the plot when the button is pressed but not when
    # the slider is changed
    @reactive.event(input.go, ignore_none=False)
    def plot():
        np.random.seed(19680801)
        x = 100 + 15 * np.random.randn(input.n())
        fig, ax = plt.subplots()
        ax.hist(x, bins=30, density=True)
        return fig


app = App(app_ui, server)



import shiny.ui
from shiny.express import input, render, ui

# `ui.hold() as x` can be used to save `x` for later output
with ui.hold() as hello_card:
    with ui.card():
        with ui.span():
            "This is a"
            ui.span(" card", style="color: red;")

hello_card

hello_card

ui.hr()


# `ui.hold()` can be used to just suppress output
with ui.hold():

    @render.code()
    def txt():
        return f"Slider value: {input.n()}"


ui.input_slider("n", "N", 1, 100, 50)

shiny.ui.output_code("txt")



from shiny.express import input, render, ui

ui.page_opts(title="Kitchen Sink: ui.input_numeric()", fillable=True)

with ui.layout_columns():
    with ui.card():
        ui.h3("Default Numeric Input")
        ui.input_numeric("default", label="Default numeric input", value=10)

        @render.code
        def default_txt():
            return str(input.default())

    with ui.card():
        ui.h3("With Min and Max")
        ui.input_numeric("min_max", "Min and Max", min=0, max=100, value=50)

        @render.code
        def min_max_txt():
            return str(input.min_max())


with ui.layout_columns():
    with ui.card():
        ui.h3("With Step")
        ui.input_numeric("step", "Step of 0.5", step=0.5, value=2.5)

        @render.code
        def step_txt():
            return str(input.step())

    with ui.card():
        ui.h3("Custom Width")
        ui.input_numeric("width", "Custom width", width="200px", value=15)

        @render.code
        def width_txt():
            return str(input.width())



from shiny.express import input, render, ui

ui.page_opts(title="Kitchen Sink: ui.input_password()", fillable=True)

with ui.layout_columns():
    with ui.card():
        ui.input_password("default", label="Default password input")

        @render.code
        def default_txt():
            return str(input.default())

    with ui.card():
        ui.input_password(
            "placeholder", "With placeholder", placeholder="Enter password"
        )

        @render.code
        def placeholder_txt():
            return str(input.placeholder())


with ui.layout_columns():
    with ui.card():
        ui.input_password("width", "Custom width", width="200px")

        @render.code
        def width_txt():
            return str(input.width())

    with ui.card():
        ui.input_password("value", "With initial value", value="secret123")

        @render.code
        def value_txt():
            return str(input.value())



import pandas as pd

from shiny.express import render, ui

df = pd.DataFrame(data={"a": [str(i) for i in range(10)]})


with ui.layout_column_wrap(width=1 / 2):
    with ui.card():
        ui.card_header("Original data")

        @render.data_frame
        def my_df():
            return render.DataGrid(data=df, filters=True, selection_mode="rows")

    with ui.card():
        ui.card_header("Selected data")

        @render.data_frame
        def selected_df():
            return my_df.data_view(selected=True)

    with ui.card():
        ui.card_header("Selected rows")

        @render.code
        def selected_rows():
            return str(my_df.cell_selection()["rows"])



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_popover_kitchensink(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    popover_auto = controller.Popover(page, "btn_popover_title")
    popover_auto.expect_active(False)
    popover_auto.set(True)
    popover_auto.expect_active(True)
    popover_auto.expect_body("Placement should be auto along with a title")
    popover_auto.expect_title("Popover title")
    popover_auto.expect_placement(
        "right"
    )  # since there is no space on the top, it defaults to right
    popover_auto.set(False)
    popover_auto.expect_active(False)

    popover_top = controller.Popover(page, "btn_popover_top")
    popover_top.expect_active(False)
    popover_top.set(True)
    popover_top.expect_active(True)
    popover_top.expect_body("Popover placement should be on the top")
    popover_top.expect_placement("top")
    popover_top.set(False)
    popover_top.expect_active(False)

    popover_right = controller.Popover(page, "btn_popover_right")
    popover_right.expect_active(False)
    popover_right.set(True)
    popover_right.expect_active(True)
    popover_right.expect_body("Popover placement should be on the right")
    popover_right.expect_placement("right")
    popover_right.set(False)
    popover_right.expect_active(False)

    popover_left = controller.Popover(page, "btn_popover_left")
    popover_left.expect_active(False)
    popover_left.set(True)
    popover_left.expect_active(True)
    popover_left.expect_body("Popover placement should be on the left")
    popover_left.expect_placement("left")
    popover_left.set(False)
    popover_left.expect_active(False)



from shiny.express import input, render, ui

with ui.card(id="card"):
    ui.input_slider("s1", "A", 1, 100, 20)

    with ui.hold():

        @render.code
        def hidden():
            return input.s1()

    ui.input_slider("s2", "B", 1, 100, 40)

    # from shiny.express import ui_kwargs
    # @ui_kwargs(placeholder=False)
    # @ui_kwargs(placeholder=True)
    with ui.hold() as held:

        @render.code()
        def visible():
            # from shiny import req

            # req(False)
            return input.s2()

    "Some text here"
    held  # pyright: ignore[reportUnusedExpression]



from __future__ import annotations

import time

from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def click_extended_task_button(
    button: controller.InputTaskButton,
    current_time: controller.OutputText,
) -> str:
    button.expect_state("ready")
    button.click(timeout=0)
    button.expect_state("busy", timeout=0)
    return current_time.get_value(timeout=0)


def test_input_action_task_button(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)
    y = controller.InputNumeric(page, "y")
    y.set("4")
    result = controller.OutputText(page, "show_result")
    current_time = controller.OutputText(page, "current_time")
    # Make sure the time has content
    current_time.expect.not_to_be_empty()

    # Wait until shiny is stable
    result.expect_value("3")

    # Extended task
    button_task = controller.InputTaskButton(page, "btn_task")
    button_task.expect_label_busy("\n  \n Processing...")
    button_task.expect_label_ready("Non-blocking task")
    button_task.expect_auto_reset(True)
    # Click button and collect the current time from the app
    time1 = click_extended_task_button(
        button_task,
        current_time,
    )
    # Make sure time value updates (before the calculation finishes
    current_time.expect.not_to_have_text(time1, timeout=500)
    result.expect_value("3", timeout=0)
    # After the calculation time plus a buffer, make sure the calculation finishes
    result.expect_value("5", timeout=(1.5 + 1) * 1000)

    # set up Blocking test
    y.set("15")
    result.expect_value("5")

    # Blocking verification
    button_block = controller.InputTaskButton(page, "btn_block")
    button_block.expect_label_busy("\n  \n Blocking...")
    button_block.expect_label_ready("Block compute")
    button_block.expect_auto_reset(True)
    time_block = click_extended_task_button(
        button_block,
        current_time,
    )
    # Make sure time value has not changed after 500ms has ellapsed
    time.sleep(0.5)
    current_time.expect_value(time_block, timeout=0)



from __future__ import annotations

import json
import typing
from typing import Literal

from playwright.sync_api import Locator, Page
from playwright.sync_api import expect as playwright_expect

from ...types import MISSING_TYPE
from .._types import AttrValue, PatternOrStr, StyleValue, Timeout
from ..expect._internal import (
    expect_attribute_to_have_value as _expect_attribute_to_have_value,
)
from ..expect._internal import expect_class_to_have_value as _expect_class_to_have_value
from ..expect._internal import expect_style_to_have_value as _expect_style_to_have_value
from ._base import (
    Resize,
    UiBaseP,
    UiWithLabel,
    WidthContainerStyleM,
    all_missing,
    not_is_missing,
    set_text,
)


class _SetTextM:
    def set(self: UiBaseP, value: str, *, timeout: Timeout = None) -> None:
        """
        Sets the text value

        Parameters
        ----------
        value
            The text to set.
        timeout
            The maximum time to wait for the text to be set. Defaults to `None`.
        """
        set_text(self.loc, value, timeout=timeout)


class _ExpectTextInputValueM:
    """A mixin class for text input values."""

    def expect_value(
        self: UiBaseP,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the value of the text input to have a specific value.

        Parameters
        ----------
        value
            The expected value of the text input.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        playwright_expect(self.loc).to_have_value(value, timeout=timeout)


class InputNumeric(
    _SetTextM,
    _ExpectTextInputValueM,
    WidthContainerStyleM,
    UiWithLabel,
):
    """Controller for :func:`shiny.ui.input_numeric`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes the input numeric Playwright control.

        Parameters
        ----------
        page
            The page where the input numeric is located.
        id
            The id of the input numeric.
        """
        super().__init__(
            page,
            id=id,
            loc=f"input#{id}[type=number].shiny-bound-input",
        )

    def expect_min(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the minimum numeric value to be a specific value.

        Parameters
        ----------
        value
            The expected minimum numeric value.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(self.loc, "min", value=value, timeout=timeout)

    def expect_max(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the maximum numeric value to be a specific value.

        Parameters
        ----------
        value
            The expected maximum numeric value.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(self.loc, "max", value=value, timeout=timeout)

    def expect_step(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect step value when incrementing/decrementing the numeric input.

        Parameters
        ----------
        value
            The expected step value for the numeric input.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(self.loc, "step", value=value, timeout=timeout)


class _ExpectSpellcheckAttrM:
    """
    A mixin class for the spellcheck attribute.
    """

    def expect_spellcheck(
        self: UiBaseP,
        value: Literal["true", "false"] | None,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the `spellcheck` attribute of the input to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `spellcheck` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        # self.spellcheck.expect_to_have_value(value, timeout=timeout)
        _expect_attribute_to_have_value(
            self.loc, "spellcheck", value=value, timeout=timeout
        )


class _ExpectPlaceholderAttrM:
    def expect_placeholder(
        self: UiBaseP,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the `placeholder` attribute of the input to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `placeholder` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "placeholder", value=value, timeout=timeout
        )


class _ExpectAutocompleteAttrM:
    def expect_autocomplete(
        self: UiBaseP,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the `autocomplete` attribute of the input to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `autocomplete` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "autocomplete", value=value, timeout=timeout
        )


class InputText(
    _SetTextM,
    _ExpectTextInputValueM,
    WidthContainerStyleM,
    _ExpectPlaceholderAttrM,
    _ExpectAutocompleteAttrM,
    _ExpectSpellcheckAttrM,
    UiWithLabel,
):
    """Controller for :func:`shiny.ui.input_text`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes the input text.

        Parameters
        ----------
        page
            The page where the input text is located.
        id
            The id of the input text.
        """
        super().__init__(
            page,
            id=id,
            loc=f"input#{id}[type=text].shiny-bound-input",
        )


class InputPassword(
    _SetTextM,
    _ExpectTextInputValueM,
    WidthContainerStyleM,
    _ExpectPlaceholderAttrM,
    UiWithLabel,
):
    """Controller for :func:`shiny.ui.input_password`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes the input password.

        Parameters
        ----------
        page
            The page where the input password is located.
        id
            The id of the input password.
        """
        super().__init__(
            page,
            id=id,
            loc=f"input#{id}[type=password].shiny-bound-input",
        )

    # This class does not inherit from `_WidthContainerM`
    # as the width is in the element style
    def expect_width(
        self,
        value: StyleValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the `width` attribute of the input password to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `width` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_style_to_have_value(self.loc_container, "width", value, timeout=timeout)


class InputTextArea(
    _SetTextM,
    WidthContainerStyleM,
    _ExpectTextInputValueM,
    _ExpectPlaceholderAttrM,
    _ExpectAutocompleteAttrM,
    _ExpectSpellcheckAttrM,
    UiWithLabel,
):
    """Controller for :func:`shiny.ui.input_text_area`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes the input text area.

        Parameters
        ----------
        page
            The page where the input text area is located.
        id
            The id of the input text area.
        """
        super().__init__(
            page,
            id=id,
            loc=f"textarea#{id}.shiny-bound-input",
        )

    def expect_width(self, value: StyleValue, *, timeout: Timeout = None) -> None:
        """
        Expect the `width` attribute of the input text area to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `width` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        if value is None:
            _expect_style_to_have_value(
                self.loc_container,
                "width",
                None,
                timeout=timeout,
            )
            _expect_style_to_have_value(self.loc, "width", "100%", timeout=timeout)
        else:
            _expect_style_to_have_value(
                self.loc_container, "width", value, timeout=timeout
            )
            _expect_style_to_have_value(self.loc, "width", None, timeout=timeout)

    def expect_height(self, value: StyleValue, *, timeout: Timeout = None) -> None:
        """
        Expect the `height` attribute of the input text area to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `height` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_style_to_have_value(self.loc, "height", value, timeout=timeout)

    def expect_cols(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        """
        Expect the `cols` attribute of the input text area to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `cols` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(self.loc, "cols", value=value, timeout=timeout)

    def expect_rows(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        """
        Expect the `rows` attribute of the input text area to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `rows` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(self.loc, "rows", value=value, timeout=timeout)

    def expect_resize(
        self,
        value: Resize | None,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the `resize` attribute of the input text area to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `resize` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_style_to_have_value(self.loc, "resize", value=value, timeout=timeout)

    def expect_autoresize(
        self,
        value: bool,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the `autoresize` attribute of the input text area to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `autoresize` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_class_to_have_value(
            self.loc,
            "textarea-autoresize",
            has_class=value,
            timeout=timeout,
        )


class _DateBase(
    WidthContainerStyleM,
    _SetTextM,
    UiWithLabel,
):

    # Due to the `language` parameter, we can't use `datetime.date` as a value type

    def expect_value(
        self,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        # if value is None:
        #     self.expect.to_be_empty(timeout=timeout)
        # else:
        self.expect.to_have_value(value, timeout=timeout)

    def expect_min_date(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected `data-min-date` attribute value.

        Parameters
        ----------
        value
            The expected `data-min-date` attribute value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-min-date", value=value, timeout=timeout
        )

    def expect_max_date(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected `data-max-date` attribute value.

        Parameters
        ----------
        value
            The expected `data-max-date` attribute value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-max-date", value=value, timeout=timeout
        )

    def expect_format(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected `data-date-format` attribute value.

        Parameters
        ----------
        value
            The expected `data-date-format` attribute value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-date-format", value=value, timeout=timeout
        )

    def expect_startview(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected `data-date-start-view` attribute value.

        Parameters
        ----------
        value
            The expected `data-date-start-view` attribute value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-date-start-view", value=value, timeout=timeout
        )

    def expect_weekstart(
        self,
        value: int | AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected `data-date-week-start` attribute value.

        Parameters
        ----------
        value
            The expected `data-date-week-start` attribute value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        if isinstance(value, int):
            value = str(value)
        _expect_attribute_to_have_value(
            self.loc, "data-date-week-start", value=value, timeout=timeout
        )

    def expect_language(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected `data-date-language` attribute value.

        Parameters
        ----------
        value
            The expected `data-date-language` attribute value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-date-language", value=value, timeout=timeout
        )

    # autoclose: bool = True,
    def expect_autoclose(
        self,
        value: Literal["true", "false"],
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected `data-date-autoclose` attribute value.

        Parameters
        ----------
        value
            The expected `data-date-autoclose` attribute value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "data-date-autoclose", value=value, timeout=timeout
        )

    def expect_datesdisabled(
        self,
        value: list[str] | None,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected `data-date-dates-disabled` attribute value.

        Parameters
        ----------
        value
            The expected `data-date-dates-disabled` attribute value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        if isinstance(value, list):
            assert len(value) > 0, "`value` must be `None` or a non-empty list"
        value_str = "null" if value is None else json.dumps(value)
        _expect_attribute_to_have_value(
            self.loc,
            "data-date-dates-disabled",
            value=value_str,
            timeout=timeout,
        )

    def expect_daysofweekdisabled(
        self,
        value: list[int] | None,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected `data-date-days-of-week-disabled` attribute value.

        Parameters
        ----------
        value
            The expected `data-date-days-of-week-disabled` attribute value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        if isinstance(value, list):
            assert len(value) > 0, "`value` must be `None` or a non-empty list"
        value_str = "null" if value is None else json.dumps(value)
        _expect_attribute_to_have_value(
            self.loc,
            "data-date-days-of-week-disabled",
            value=value_str,
            timeout=timeout,
        )


class InputDate(_DateBase):
    def __init__(self, page: Page, id: str) -> None:
        """
        Initialize an InputDate object.

        Parameters
        ----------
        page
            The page object.
        id
            The id of the input element.
        """
        super().__init__(
            page,
            id=id,
            loc="input[type=text].form-control",
            loc_container=f"div#{id}.shiny-input-container",
        )


class InputDateRange(WidthContainerStyleM, UiWithLabel):
    """Controller for :func:`shiny.ui.input_date_range`."""

    loc_separator: Locator
    """
    Playwright `Locator` of the separator between the two input elements.
    """
    loc_start: Locator
    """
    Playwright `Locator` of the start date input element.
    """
    loc_end: Locator
    """
    Playwright `Locator` of the end date input element.
    """
    date_start: _DateBase
    """
    The start date input element.
    """
    date_end: _DateBase
    """
    The end date input element.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initialize an InputDateRange object.

        Parameters
        ----------
        page
            The page object.
        id
            The id of the input element.
        """
        super().__init__(
            page,
            id=id,
            loc="input[type=text].form-control",
            loc_container=f"div#{id}.shiny-input-container",
        )
        self.loc_separator = self.loc_container.locator(".input-daterange > span")
        self.loc_start = self.loc.nth(0)
        self.loc_end = self.loc.nth(1)
        self.date_start = _DateBase(
            page,
            id=id,
            loc=self.loc_start,
            loc_label=self.loc_label,
            loc_container=self.loc_container,
        )
        self.date_end = _DateBase(
            page,
            id=id,
            loc=self.loc_end,
            loc_label=self.loc_label,
            loc_container=self.loc_container,
        )

    def set(
        self,
        value: typing.Tuple[
            str | None,
            str | None,
        ],
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Sets the value of the input element.

        Parameters
        ----------
        value
            The value to set. The first element is the start date and the second element is the end date.
        timeout
            The maximum time to wait for the value to be set. Defaults to `None`.
        """
        start = value[0]
        end = value[1]
        # TODO-future; Composable set() methods?
        if start is not None:
            self.date_start.set(value=start, timeout=timeout)
        if end is not None:
            self.date_end.set(value=end, timeout=timeout)

    def expect_value(
        self,
        value: (
            typing.Tuple[PatternOrStr, PatternOrStr]
            | typing.Tuple[PatternOrStr, MISSING_TYPE]
            | typing.Tuple[MISSING_TYPE, PatternOrStr]
        ),
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected value.

        Parameters
        ----------
        value
            The expected value. The first element is the start date and the second element is the end date.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        if all_missing(*value):
            raise ValueError("Both `start_val` and `end_val` can not be `MISSING_TYPE`")
        start_val = value[0]
        end_val = value[1]
        # We can not use `[value={value}]` within Locators.
        # The physical `value` attribute is never set, so we can not select on it.
        # We must as the start and end values individually, rather than at the same time like the checkboxgroup input.
        # TODO-future; Composable expectations
        if not_is_missing(start_val):
            self.date_start.expect_value(start_val, timeout=timeout)
        if not_is_missing(end_val):
            self.date_end.expect_value(end_val, timeout=timeout)

    # min: Optional[Union[date, str]] = None,
    def expect_min_date(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected minimum date.

        Parameters
        ----------
        value
            The expected minimum date.
        timeout
            The maximum time to wait for the minimum date to appear. Defaults to `None`.
        """
        # TODO-future; Composable expectations
        self.date_start.expect_min_date(value, timeout=timeout)
        self.date_end.expect_min_date(value, timeout=timeout)

    # max: Optional[Union[date, str]] = None,
    def expect_max_date(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected maximum date.

        Parameters
        ----------
        value
            The expected maximum date.
        timeout
            The maximum time to wait for the maximum date to appear. Defaults to `None`.
        """
        # TODO-future; Composable expectations
        self.date_start.expect_max_date(value, timeout=timeout)
        self.date_end.expect_max_date(value, timeout=timeout)

    # format: str = "yyyy-mm-dd",
    def expect_format(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected format.

        Parameters
        ----------
        value
            The expected format.
        timeout
            The maximum time to wait for the format to appear. Defaults to `None`.
        """
        # TODO-future; Composable expectations
        self.date_start.expect_format(value, timeout=timeout)
        self.date_end.expect_format(value, timeout=timeout)

    # startview: str = "month",
    def expect_startview(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected start view.

        Parameters
        ----------
        value
            The expected start view.
        timeout
            The maximum time to wait for the start view to appear. Defaults to `None`.
        """
        # TODO-future; Composable expectations
        self.date_start.expect_startview(value, timeout=timeout)
        self.date_end.expect_startview(value, timeout=timeout)

    # weekstart: int = 0,
    def expect_weekstart(
        self,
        value: int | AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected week start.

        Parameters
        ----------
        value
            The expected week start.
        timeout
            The maximum time to wait for the week start to appear. Defaults to `None`.
        """
        # TODO-future; Composable expectations
        self.date_start.expect_weekstart(value, timeout=timeout)
        self.date_end.expect_weekstart(value, timeout=timeout)

    # language: str = "en",
    def expect_language(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected language.

        Parameters
        ----------
        value
            The expected language.
        timeout
            The maximum time to wait for the language to appear. Defaults to `None`.
        """
        # TODO-future; Composable expectations
        self.date_start.expect_language(value, timeout=timeout)
        self.date_end.expect_language(value, timeout=timeout)

    # separator: str = " to ",
    def expect_separator(
        self,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected separator.

        Parameters
        ----------
        value
            The expected separator.
        timeout
            The maximum time to wait for the separator to appear. Defaults to `None`.
        """
        playwright_expect(self.loc_separator).to_have_text(value, timeout=timeout)

    # autoclose: bool = True,
    def expect_autoclose(
        self,
        value: Literal["true", "false"],
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the input element has the expected autoclose value.

        Parameters
        ----------
        value
            The expected autoclose value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        # TODO-future; Composable expectations
        self.date_start.expect_autoclose(value, timeout=timeout)
        self.date_end.expect_autoclose(value, timeout=timeout)



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_input_selectize_kitchensink(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    basic_selectize = controller.InputSelectize(page, "basic_selectize")
    basic_select_txt = controller.OutputText(page, "basic_result_txt")
    # # TODO-karan; Debug why this does not complete
    # basic_selectize.expect_choices(["Apple", "Banana", "Cherry", "Date", "Elderberry"])
    basic_selectize.expect_choice_labels(
        [
            "Apple",
            "Banana",
            "Cherry",
            "Date",
            "Elderberry",
        ]
    )
    basic_selectize.expect_choice_groups([])
    basic_selectize.expect_label("Default selectize")
    basic_select_txt.expect_value("Apple")
    basic_selectize.expect_multiple(False)

    selectize_with_label = controller.InputSelectize(page, "selectize_with_label")
    selectize_with_label_txt = controller.OutputText(page, "selectize_with_label_txt")
    # # TODO-karan; Debug why this does not complete
    # selectize_with_label.expect_choices(
    #     ["apple", "banana", "cherry", "date", "elderberry"]
    # )
    selectize_with_label.expect_choice_labels(
        [
            "Apple",
            "Banana",
            "Cherry",
            "Date",
            "Elderberry",
        ]
    )
    selectize_with_label.expect_choice_groups([])
    selectize_with_label_txt.expect_value("apple")

    multiple_selectize = controller.InputSelectize(page, "multi_selectize")
    multiple_selectize_txt = controller.OutputText(page, "multi_result_txt")
    multiple_options = ["Banana", "Cherry"]
    multiple_selectize.set(multiple_options)
    multiple_selectize.expect_selected(["Banana", "Cherry"])
    multiple_selectize_txt.expect_value("Banana, Cherry")
    for option in multiple_options:
        # click on the remove button for each selected option
        multiple_selectize.loc.locator(
            f"+ div.plugin-remove_button > .has-options > .item[data-value={option}] > .remove"
        ).click()
        page.keyboard.press(
            "Escape"
        )  # to remove dropdown from blocking access to other selectize inputs
    multiple_selectize_txt.expect_value("")
    multiple_selectize.expect_multiple(True)

    selectize_with_selected = controller.InputSelectize(page, "selectize_with_selected")
    selectize_with_selected_txt = controller.OutputText(page, "selected_result_txt")
    selectize_with_selected.expect_selected(["Cherry"])
    selectize_with_selected_txt.expect_value("Cherry")

    selectize_width_close_button = controller.InputSelectize(
        page, "selectize_width_close_button"
    )
    selectize_width_close_button_txt = controller.OutputText(
        page, "selectize_width_close_button_txt"
    )
    selectize_width_close_button_txt.expect_value("Orange")
    selectize_width_close_button.expect_width("400px")
    selectize_width_close_button.expect_choice_groups(["Citrus", "Berries"])
    # # TODO-karan; Debug why this does not complete
    # selectize_width_close_button.expect_choices(
    #     ["Orange", "Lemon", "Lime", "Strawberry", "Blueberry", "Raspberry"]
    # )
    selectize_width_close_button.expect_choice_labels(
        [
            "Sweet and tangy",
            "Zesty and refreshing",
            "Bright and tart",
            "Juicy and sweet",
            "Tiny and antioxidant-rich",
            "Delicate and slightly tart",
        ]
    )
    selectize_width_close_button.loc.locator("..").locator(
        "> div.plugin-clear_button > a.clear"
    ).click()  # Clear default selection
    selectize_width_close_button_txt.expect_value("")  # Expecting empty after clear



from shiny.express import ui

ui.input_checkbox("show", "Show radio buttons", False)

with ui.panel_conditional("input.show"):
    ui.input_radio_buttons("radio", "Choose ", ["slider", "select"])

with ui.panel_conditional("input.show && input.radio === 'slider'"):
    ui.input_slider("slider", None, min=0, max=100, value=50)

with ui.panel_conditional("input.show && input.radio === 'select'"):
    ui.input_select("select", None, ["A", "B", "C"])



import sys

from htmltools import Tag, Tagifiable, TagList


# A decorator used for `def` statements. It makes sure that any `def` statement which
# returns a tag-like object, or one with a `_repr_html` method, will be passed on to
# the current sys.displayhook.
def _expressify_decorator_function_def(fn: object) -> object:
    if isinstance(fn, (Tag, TagList, Tagifiable)) or hasattr(fn, "_repr_html_"):
        sys.displayhook(fn)

    return fn



from datetime import datetime

from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.input_action_button("flush", "Trigger flush"),
    ui.output_ui("n_clicks"),
    ui.div(id="flush_time"),
)


def server(input: Inputs, output: Outputs, session: Session):
    def log():
        msg = "A reactive flush occurred at " + datetime.now().strftime("%H:%M:%S:%f")
        print(msg)
        ui.insert_ui(
            ui.p(msg),
            selector="#flush_time",
        )

    session.on_flush(log, once=False)

    @render.ui
    def n_clicks():
        return "Number of clicks: " + str(input.flush())


app = App(app_ui, server)



from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("output_text")


def test_output_text_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    txt = controller.InputText(page, "txt")
    text = controller.OutputText(page, "text")
    verb = controller.OutputTextVerbatim(page, "verb")
    verb_no_placeholder = controller.OutputTextVerbatim(page, "verb_no_placeholder")

    txt.set("")  # Reset text

    text.expect_value("")
    text.expect_inline(False)

    verb.expect_value("")
    verb.expect_has_placeholder(True)

    verb_no_placeholder.expect_value("")
    verb_no_placeholder.expect_has_placeholder(False)

    txt_val = "test value 42"
    txt.set(txt_val)

    text.expect_value(txt_val)
    verb.expect_value(txt_val)
    verb_no_placeholder.expect_value(txt_val)



from typing import Optional

import ipyleaflet as L
from shinywidgets import output_widget, reactive_read, register_widget

from shiny import Inputs, Outputs, Session, module, reactive, req, ui

# ============================================================
# Module: location
# ============================================================


@module.ui
def location_ui(
    label: str = "Location",
    *,
    lat: Optional[float] = None,
    long: Optional[float] = None,
) -> ui.TagChild:
    return ui.div(
        ui.input_numeric("lat", "Latitude", value=lat),
        ui.input_numeric("long", "Longitude", value=long),
        ui.help_text("Click to select location"),
        output_widget("map", height="200px"),
        ui.tags.style(
            """
            .jupyter-widgets.leaflet-widgets {
                height: 100% !important;
            }
            """
        ),
    )


@module.server
def location_server(
    input: Inputs, output: Outputs, session: Session, *, wrap_long: bool = True
):
    map = L.Map(center=(0, 0), zoom=1, scoll_wheel_zoom=True)
    with reactive.isolate():
        marker = L.Marker(location=(input.lat() or 0, input.long() or 0))

    with reactive.isolate():  # Use this to ensure we only execute one time
        if input.lat() is None and input.long() is None:
            ui.notification_show(
                "Searching for location...", duration=99999, id="searching"
            )
            ui.insert_ui(
                ui.tags.script(
                    """
                    navigator.geolocation.getCurrentPosition(
                        ({coords}) => {
                            const {latitude, longitude, altitude} = coords;
                            Shiny.setInputValue("#HERE#", {latitude, longitude});
                        },
                        (err) => {
                            Shiny.setInputValue("#HERE#", {latitude: 0, longitude: 0});
                        },
                        {maximumAge: Infinity, timeout: Infinity}
                    )
                    """.replace(
                        "#HERE#", module.resolve_id("here")
                    )
                ),
                selector="body",
                where="beforeEnd",
                immediate=True,
            )

    @reactive.isolate()
    def update_text_inputs(lat: Optional[float], long: Optional[float]) -> None:
        req(lat is not None, long is not None)
        lat = round(lat, 8)
        long = round(long, 8)
        if lat != input.lat():
            input.lat.freeze()
            ui.update_text("lat", value=lat)
        if long != input.long():
            input.long.freeze()
            ui.update_text("long", value=long)
        map.center = (lat, long)

    @reactive.isolate()
    def update_marker(lat: Optional[float], long: Optional[float]) -> None:
        req(lat is not None, long is not None)
        lat = round(lat, 8)
        long = round(long, 8)
        if marker.location != (lat, long):
            marker.location = (lat, long)
        if marker not in map.layers:
            map.add_layer(marker)
        map.center = marker.location

    def on_map_interaction(**kwargs):
        if kwargs.get("type") == "click":
            lat, long = kwargs.get("coordinates")
            update_text_inputs(lat, long)

    map.on_interaction(on_map_interaction)

    register_widget("map", map)

    @reactive.effect
    def _():
        coords = reactive_read(marker, "location")
        if coords:
            update_text_inputs(coords[0], coords[1])

    @reactive.effect
    def sync_autolocate():
        coords = input.here()
        ui.notification_remove("searching")
        if coords and not input.lat() and not input.long():
            update_text_inputs(coords["latitude"], coords["longitude"])

    @reactive.effect
    def sync_inputs_to_marker():
        update_marker(input.lat(), input.long())

    @reactive.calc
    def location():
        """Returns tuple of (lat,long) floats--or throws silent error if no lat/long is
        selected"""

        # Require lat/long to be populated before we can proceed
        req(input.lat() is not None, input.long() is not None)

        try:
            long = input.long()
            # Wrap longitudes so they're within [-180, 180]
            if wrap_long:
                long = (long + 180) % 360 - 180
            return (input.lat(), long)
        except ValueError:
            raise ValueError("Invalid latitude/longitude specification")

    return location



from icons import gear_fill

from shiny.express import input, render, ui

with ui.popover(id="btn_popover"):
    ui.input_action_button("btn", "A button", class_="mt-3")

    "A popover with more context and information than should be used in a tooltip."
    "You can even have multiple DOM elements in a popover!"


with ui.card(class_="mt-3"):
    with ui.card_header():
        "Plot title (Click the gear to change variables)"
        with ui.popover(placement="right", id="card_popover"):
            ui.span(gear_fill, style="position:absolute; top: 5px; right: 7px;")
            "Put dropdowns here to alter your plot!"
            ui.input_selectize("x", "X", ["x1", "x2", "x3"])
            ui.input_selectize("y", "Y", ["y1", "y2", "y3"])

    @render.text
    def plot_txt():
        return f"<Making plot using x: {input.x()} and y: {input.y()}>"



from __future__ import annotations

from pathlib import Path

from ._run import wrap_express_app
from ._utils import unescape_from_var_name


# If someone requests shiny.express.app:_2f_path_2f_to_2f_app_2e_py, then we will call
# wrap_express_app(Path("/path/to/app.py")) and return the result.
def __getattr__(name: str) -> object:
    name = unescape_from_var_name(name)
    return wrap_express_app(Path(name))



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Generic,
    Optional,
    TypeVar,
    Union,
    cast,
)

from htmltools import MetadataNode, Tag, TagList

from ..._docstring import add_example
from ..._typing_extensions import Self
from ..._utils import is_async_callable, wrap_async
from ...types import Jsonifiable

if TYPE_CHECKING:
    from ...session import Session

# TODO-barret-docs: Double check docs are rendererd
# Missing first paragraph from some classes: Example: TransformerMetadata.
# No init method for TransformerParams. This is because the `DocClass` object does not
# display methods that start with `_`. Therefore no `__init__` or `__call__` methods are
# displayed. Even if they have docs.


__all__ = (
    "Renderer",
    "Jsonifiable",
    "ValueFn",
    "Jsonifiable",
    "AsyncValueFn",
    "RendererT",
)


RendererT = TypeVar("RendererT", bound="Renderer[Any]")
"""
Generic output renderer class to pass the original Renderer subclass through a decorator
function.

When accepting and returning a `Renderer` class, utilize this TypeVar as to not reduce
the variable type to `Renderer[Any]`
"""

IT = TypeVar("IT")
"""
Return type from the user-supplied value function passed into the renderer.
"""


DefaultUIFnResult = Union[TagList, Tag, MetadataNode, str]
DefaultUIFnResultOrNone = Union[DefaultUIFnResult, None]
DefaultUIFn = Callable[[str], DefaultUIFnResultOrNone]

# Requiring `None` type throughout the value functions as `return` returns `None` type.
# This is typically paired with `req(False)` to exit quickly.
# If package authors want to NOT allow `None` type, they can capture it in a custom
#   render method with a runtime error. (Or make a new RendererThatCantBeNone class)
ValueFn = Union[
    Callable[[], Union[IT, None]],
    Callable[[], Awaitable[Union[IT, None]]],
]
"""
App-supplied output value function which returns type `IT` or `None`. This function can
be synchronous or asynchronous.
"""


@add_example()
class Renderer(Generic[IT]):
    """
    Output renderer class

    An output renderer is a class that will take in a callable function (value
    function), transform the returned value into a JSON-serializable object, and send
    the result to the browser.

    When the value function is received, the renderer will be auto registered with
    the current session's `Output` class, hooking it into Shiny's reactive graph. By
    auto registering as an `Output`, it allows for App authors to skip adding `@output`
    above the renderer. (If programmatic `id` is needed, `@output(id="foo")` can still be
    used!)

    There are two methods that must be implemented by the subclasses:
    `.auto_output_ui(self)` and either `.transform(self, value: IT)` or `.render(self)`.

    * In Express mode, the output renderer will automatically render its UI via
      `.auto_output_ui(self)`. This helper method allows App authors to skip adding a
      `ui.output_*` function to their UI, making Express mode even more concise. If more
      control is needed over the UI, `@ui.hold` can be used to suppress the auto
      rendering of the UI. When using `@ui.hold` on a renderer, the renderer's UI will
      need to be added to the app to connect the rendered output to Shiny's reactive
      graph.
    * The `render` method is responsible for executing the value function and performing
      any transformations for the output value to be JSON-serializable (`None` is a
      valid value!). To avoid the boilerplate of resolving the value function and
      returning early if `None` is received, package authors may implement the
      `.transform(self, value: IT)` method. The `transform` method's sole job is to
      _transform_ non-`None` values into an object that is JSON-serializable.
    """

    # Q: Could we do this with typing without putting `P` in the Generic?
    # A: No. Even if we had a `P` in the Generic, the calling decorator would not have access to it.
    # Idea: Possibly use a chained method of `.ui_kwargs()`? https://github.com/posit-dev/py-shiny/issues/971
    _auto_output_ui_kwargs: dict[str, Any] = dict()

    __name__: str
    """
    Name of output function supplied. (The value **will not** contain a module prefix.)

    Set within `.__call__()` method.
    """

    output_id: str
    """
    Output function name or ID (provided to `@output(id=)`).

    This value **will not** contain a module prefix (or session name-spacing). To get
    the fully resolved ID, call
    `shiny.session.require_active_session(None).ns(self.output_id)`.

    An initial value of `.__name__` (set within `Renderer.__call__(_fn)`) will be used
    until the output renderer is registered within the session.
    """

    fn: AsyncValueFn[IT]
    """
    App-supplied output value function which returns type `IT`. This function is always
    asyncronous as the original app-supplied function possibly wrapped to execute
    asynchonously.
    """

    def __call__(self, _fn: ValueFn[IT]) -> Self:
        """
        Add the value function to the renderer.

        Addition actions performed:
        * Store the value function name.
        * Set the Renderer's `output_id` to the function name.
        * Auto register (given a `Session` exists) the Renderer

        Parameters
        ----------
        _fn
            Value function supplied by the App author. This function can be synchronous
            or asynchronous.

        Returns
        -------
        :
            Original renderer instance.
        """
        from ...session import get_current_session

        if not callable(_fn):
            raise TypeError("Value function must be callable")

        # Set value function with extra meta information
        self.fn = AsyncValueFn(_fn)

        # Copy over function name as it is consistent with how Session and Output
        # retrieve function names
        self.__name__ = _fn.__name__

        # Set the value of `output_id` to the function name.
        # This helps with testing and other situations where no session is present
        # for auto-registration to occur.
        self.output_id = self.__name__

        self._session = get_current_session()

        # Allow for App authors to not require `@output`
        self._auto_register()

        # Return self for possible chaining of methods!
        return self

    def _set_output_metadata(
        self,
        *,
        output_id: str,
    ) -> None:
        """
        Method to be called within `@output` to set the renderer's metadata.

        Parameters
        ----------
        output_id : str
            Output function name or ID (provided to `@output(id=)`). This value **will
            not** contain a module prefix (or session name-spacing).
        """
        self.output_id = output_id

    def auto_output_ui(
        self,
        # *
        # **kwargs: object,
    ) -> DefaultUIFnResultOrNone:
        """
        Express mode method that automatically generates the output's UI.
        """
        return None

    def __init__(
        self,
        _fn: Optional[ValueFn[IT]] = None,
    ) -> None:
        # Do not display docs here. If docs are present, it could highjack the docs of
        # the subclass's `__init__` method.
        # """
        # Renderer - init docs here
        # """
        self._session: Session | None = None
        super().__init__()

        self._auto_registered: bool = False

        # Must be done last
        if callable(_fn):
            # Register the value function
            self(_fn)

    async def transform(self, value: IT) -> Jsonifiable:
        """
        Transform an output value into a JSON-serializable object.

        When subclassing `Renderer`, this method can be implemented to transform
        non-`None` values into a JSON-serializable object.

        If a `.render()` method is not implemented, this method **must** be implemented.
        When the output is requested, the `Renderer`'s `.render()` method will execute
        the output value function, return `None` if the value is `None`, and call this
        method to transform the value into a JSON-serializable object.

        Note, only one of `.transform()` or `.render()` should be implemented.
        """
        raise NotImplementedError(
            "Please implement either the `transform(self, value: IT)`"
            " or `render(self)` method.\n"
            "* `transform(self, value: IT)` should transform the non-`None` `value`"
            " (of type `IT`) into a JSON-serializable object."
            " Ex: `dict`, `None`, `str`. (common)\n"
            "* `render(self)` method has full control of how an App author's value is"
            " retrieved (`self._fn()`) and processed. (rare)"
        )

    async def render(self) -> Jsonifiable:
        """
        Renders the output value function.

        This method is called when the renderer is requested to render its output.

        The `Renderer`'s `render()` implementation goes as follows:

        * Execute the value function supplied to the renderer.
        * If the output value is `None`, `None` will be returned.
        * If the output value is not `None`, the `.transform()` method will be called to
          transform the value into a JSON-serializable object.

        When overwriting this method in a subclass, the implementation should execute
        the value function `.fn` and return the transformed value (which is
        JSON-serializable).
        """

        value = await self.fn()
        if value is None:
            return None

        rendered = await self.transform(value)
        return rendered

    # ######
    # Tagify-like methods
    # ######
    def _repr_html_(self) -> str | None:
        rendered_ui = self._render_auto_output_ui()
        if rendered_ui is None:
            return None
        return TagList(rendered_ui)._repr_html_()

    def tagify(self) -> DefaultUIFnResult:
        rendered_ui = self._render_auto_output_ui()
        if rendered_ui is None:
            raise TypeError(
                "No output UI exists for this type of render function: ",
                self.__class__.__name__,
            )
        return rendered_ui

    def _render_auto_output_ui(self) -> DefaultUIFnResultOrNone:
        from ...session import session_context

        with session_context(self._session):
            return self.auto_output_ui(
                # Pass the `@output_args(foo="bar")` kwargs through to the auto_output_ui function.
                **self._auto_output_ui_kwargs,
            )

    # ######
    # Auto registering output
    # ######

    def _on_register(self) -> None:
        if self._auto_registered:
            # We're being explicitly registered now. Undo the auto-registration.
            # (w/ module support)
            from ...session import require_active_session

            session = require_active_session(None)
            ns_name = session.output._ns(self.__name__)
            session.output.remove(ns_name)
            self._auto_registered = False
            self._session = session

    def _auto_register(self) -> None:
        """
        Auto registers the rendering method to the session output then the renderer is
        called.

        When `@output` is called on the renderer, the renderer is automatically
        un-registered via `._on_register()`.
        """
        # If in Express mode, register the output
        if not self._auto_registered:
            if self._session is not None:
                # Register output on reactive graph
                self._session.output(self)
                # We mark the fact that we're auto-registered so that, if an explicit
                # registration now occurs, we can undo this auto-registration.
                self._auto_registered = True


# Not inheriting from `WrapAsync[[], IT]` as python 3.8 needs typing extensions that
# doesn't support `[]` for a ParamSpec definition. :-( Would be minimal/clean if we
# could do `class AsyncValueFn(WrapAsync[[], IT]):`
class AsyncValueFn(Generic[IT]):
    """
    App-supplied output value function which returns type `IT`.
    asynchronous.

    Type definition: `Callable[[], Awaitable[IT]]`
    """

    def __init__(
        self,
        fn: Callable[[], IT | None] | Callable[[], Awaitable[IT | None]],
    ):
        if isinstance(fn, AsyncValueFn):
            raise TypeError(
                "Must not call `AsyncValueFn.__init__` with an object of class `AsyncValueFn`"
            )
        self._is_async = is_async_callable(fn)
        self._fn = wrap_async(fn)
        self._orig_fn = fn

    async def __call__(self) -> IT | None:
        """
        Call the asynchronous function.
        """
        return await self._fn()

    def is_async(self) -> bool:
        """
        Was the original function asynchronous?

        Returns
        -------
        :
            Whether the original function is asynchronous.
        """
        return self._is_async

    def get_async_fn(self) -> Callable[[], Awaitable[IT | None]]:
        """
        Return the async value function.

        Returns
        -------
        :
            Async wrapped value function supplied to the `AsyncValueFn` constructor.
        """
        return self._fn

    def get_sync_fn(self) -> Callable[[], IT | None]:
        """
        Retrieve the original, synchronous value function function.

        If the original function was asynchronous, a runtime error will be thrown.

        Returns
        -------
        :
            Original, synchronous function supplied to the `AsyncValueFn` constructor.
        """
        if self._is_async:
            raise RuntimeError(
                "The original function was asynchronous. Use `async_fn` instead."
            )
        sync_fn = cast(Callable[[], IT], self._orig_fn)
        return sync_fn



from __future__ import annotations

import datetime
from typing import Any, Optional, Union

from shiny import App, Inputs, Outputs, Session, module, reactive, render, ui

start_time = datetime.datetime(2023, 7, 1, 0, 0, 0, 0, tzinfo=datetime.timezone.utc)
end_time = start_time + datetime.timedelta(hours=1)


@module.ui
def slider_with_reset_ui(
    label: str,
    value: Union[
        datetime.datetime,
        tuple[datetime.datetime, datetime.datetime],
        list[datetime.datetime],
    ],
) -> ui.TagChild:
    return ui.card(
        ui.input_slider(
            "times",
            "Times",
            start_time,
            end_time,
            value,
            timezone="UTC",
            time_format="%H:%M:%S",
        ),
        ui.output_text_verbatim("txt"),
        ui.div(ui.input_action_button("reset", label)),
    )


@module.server
def slider_with_reset_server(
    input: Inputs,
    output: Outputs,
    session: Session,
    *,
    min: Optional[datetime.datetime] = None,
    max: Optional[datetime.datetime] = None,
    value: Any = None,
):
    @render.text
    def txt():
        if isinstance(input.times(), (tuple, list)):
            return " - ".join([str(x) for x in input.times()])
        else:
            return input.times()

    @reactive.effect
    @reactive.event(input.reset)
    def reset_time():
        ui.update_slider("times", min=min, max=max, value=value)


app_ui = ui.page_fluid(
    ui.layout_column_wrap(
        "400px",
        slider_with_reset_ui("one", "Jump to end", start_time),
        slider_with_reset_ui("two", "Select all", (start_time, start_time)),
        slider_with_reset_ui("three", "Select all", [start_time, start_time]),
        slider_with_reset_ui("four", "Extend min", [start_time, end_time]),
        slider_with_reset_ui("five", "Extend max", [start_time, end_time]),
        slider_with_reset_ui("six", "Extend min and max", [start_time, end_time]),
    ),
    class_="p-3",
)


def server(input: Inputs, output: Outputs, session: Session):
    slider_with_reset_server("one", value=end_time)
    slider_with_reset_server("two", value=(start_time, end_time))
    slider_with_reset_server("three", value=[start_time, end_time])
    slider_with_reset_server("four", min=start_time - datetime.timedelta(hours=1))
    slider_with_reset_server("five", max=end_time + datetime.timedelta(hours=1))
    slider_with_reset_server(
        "six",
        min=start_time - datetime.timedelta(hours=1),
        max=end_time + datetime.timedelta(hours=1),
    )


app = App(app_ui, server)



from playwright.sync_api import Page, expect
from utils.deploy_utils import skip_on_webkit

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@skip_on_webkit
def test_validate_chat_basic_error(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    chat = controller.Chat(page, "chat")

    controller.OutputCode(page, "message_state").expect.not_to_have_text(
        "",
        timeout=30 * 1000,
    )

    expect(chat.loc).to_be_visible(timeout=30 * 1000)
    chat.set_user_input("Hello!")
    chat.send_user_input()
    chat.expect_latest_message("Hello!", timeout=30 * 1000)

    error_loc = page.locator(".shiny-notification-error")
    expect(error_loc).to_be_visible()



import os
import sys
import typing
from pathlib import PurePath
from typing import Literal

from playwright.sync_api import ConsoleMessage, Page

from shiny.run import ShinyAppProc, run_shiny_app

here_tests_e2e_examples = PurePath(__file__).parent
pyshiny_root = here_tests_e2e_examples.parent.parent.parent

is_interactive = hasattr(sys, "ps1")
reruns = 1 if is_interactive else 3
reruns_delay = 0


def get_apps(path: str) -> typing.List[str]:
    full_path = pyshiny_root / path
    app_paths: typing.List[str] = []
    for folder in os.listdir(full_path):
        folder_path = os.path.join(full_path, folder)
        if os.path.isdir(folder_path):
            folder_files = os.listdir(folder_path)
            for file in folder_files:
                if os.path.isdir(os.path.join(folder_path, file)):
                    continue
                if not file.endswith(".py"):
                    continue
                if file == "app.py" or file.startswith("app-"):
                    # Return relative app path
                    app_paths.append(os.path.join(path, folder, file))
    return app_paths


app_idle_wait = {"duration": 300, "timeout": 30 * 1000}
app_hard_wait: typing.Dict[str, int] = {
    "examples/brownian": 250,
    "examples/ui-func": 250,
}
output_transformer_errors = [
    "ShinyDeprecationWarning: `shiny.render.transformer.output_transformer()`",
    "  super().__init__(",
    "  return OutputRenderer",
    # brownian example app
    "ShinyDeprecationWarning:",
    "shiny.render.transformer.output_transformer()",
]
express_warnings = ["Detected Shiny Express app. "]
app_allow_shiny_errors: typing.Dict[
    str, typing.Union[Literal[True], typing.List[str]]
] = {
    "api-examples/SafeException": True,
    "examples/global_pyplot": True,
    "examples/static_plots": [
        # acceptable warning
        "PlotnineWarning: Smoothing requires 2 or more points",
        "RuntimeWarning: divide by zero encountered",
        "UserWarning: This figure includes Axes that are not compatible with tight_layout",
    ],
    # Remove after shinywidgets accepts `Renderer` PR
    "api-examples/data_frame": [*output_transformer_errors],
    "api-examples/output_transformer": [*output_transformer_errors],
    "api-examples/render_express": [*express_warnings],
    "app-templates/multi-page": [*output_transformer_errors],
    "examples/airmass": [*output_transformer_errors],
    "examples/brownian": [*output_transformer_errors],
    "examples/model-score": [*output_transformer_errors],
    "deploys/plotly": [*output_transformer_errors],
}
app_allow_external_errors: typing.List[str] = [
    # TODO-garrick-future: Remove after fixing sidebar max_height_mobile warning
    "UserWarning: The `shiny.ui.sidebar(max_height_mobile=)`",
    "res = self.fn(*self.args, **self.kwargs)",
    # pandas >= 2.2.0
    # https://github.com/pandas-dev/pandas/blame/5740667a55aabffc660936079268cee2f2800225/pandas/core/groupby/groupby.py#L1129
    "FutureWarning: When grouping with a length-1 list-like",
    "sf: grouped.get_group",  # continutation of line above
    "FutureWarning:",
    "When grouping with a length-1 list-like",  # continutation of line above
    "data_subset = grouped_data.get_group(pd_key)",  # continutation of line above
    # plotnine: https://github.com/has2k1/plotnine/issues/713
    # mizani: https://github.com/has2k1/mizani/issues/34
    # seaborn: https://github.com/mwaskom/seaborn/issues/3457
    "FutureWarning: is_categorical_dtype is deprecated",
    "if pd.api.types.is_categorical_dtype(vector",  # continutation of line above
    # plotnine: https://github.com/has2k1/plotnine/issues/713#issuecomment-1701363058
    "FutureWarning: The default of observed=False is deprecated",
    # seaborn: https://github.com/mwaskom/seaborn/pull/3355
    "FutureWarning: use_inf_as_na option is deprecated",
    "pd.option_context('mode.use_inf_as_na",  # continutation of line above
]
app_allow_js_errors: typing.Dict[str, typing.List[str]] = {
    "examples/brownian": ["Failed to acquire camera feed:"],
}


# Altered from `shinytest2:::app_wait_for_idle()`
# https://github.com/rstudio/shinytest2/blob/b8fdce681597e9610fc078aa6e376134c404f3bd/R/app-driver-wait.R
def wait_for_idle_js(duration: int = 500, timeout: int = 30 * 1000) -> str:
    return """
        let duration = %s; // time needed to be idle
        let timeout = %s; // max total time
        console.log("Making promise to run");
        new Promise((resolve, reject) => {
            console.log('Waiting for Shiny to be stable');
            const cleanup = () => {
                $(document).off('shiny:busy', busyFn);
                $(document).off('shiny:idle', idleFn);
                clearTimeout(timeoutId);
                clearTimeout(idleId);
            }
            let timeoutId = setTimeout(() => {
                cleanup();
                reject('Shiny did not become stable within ' + timeout + 'ms');
            }, +timeout); // make sure timeout is number
            let idleId = null;
            const busyFn = () => {
                // clear timeout. Calling with `null` is ok.
                clearTimeout(idleId);
            };
            const idleFn = () => {
                const fn = () => {
                    // Made it through the required duration
                    // Remove event listeners
                    cleanup();
                    console.log('Shiny has been idle for ' + duration + 'ms');
                    // Resolve the promise
                    resolve();
                };
                // delay the callback wrapper function
                idleId = setTimeout(fn, +duration);
            };
            // set up individual listeners for this function.
            $(document).on('shiny:busy', busyFn);
            $(document).on('shiny:idle', idleFn);
            // if already idle, call `idleFn` to kick things off.
            if (! $("html").hasClass("shiny-busy")) {
                idleFn();
            }
        })
    """ % (
        duration,
        timeout,
    )


def wait_for_idle_app(
    page: Page, duration: int = 500, timeout: int = 30 * 1000
) -> None:
    page.evaluate(
        wait_for_idle_js(duration, timeout),
    )


def validate_example(page: Page, ex_app_path: str) -> None:
    sa: ShinyAppProc = run_shiny_app(pyshiny_root / ex_app_path, wait_for_start=True)

    console_errors: typing.List[str] = []

    def on_console_msg(msg: ConsoleMessage) -> None:
        if msg.type == "error":
            # Do not report missing favicon errors
            if msg.location["url"].endswith("favicon.ico"):
                return
            console_errors.append(msg.text)

    page.on("console", on_console_msg)

    # Makes sure the app is closed when exiting the code block
    with sa:
        page.goto(sa.url)

        app_name = os.path.basename(os.path.dirname(ex_app_path))
        short_app_path = f"{os.path.basename(os.path.dirname(os.path.dirname(ex_app_path)))}/{app_name}"

        if short_app_path in app_hard_wait.keys():
            # Apps are constantly invalidating and will not stabilize
            # Instead, wait for specific amount of time
            page.wait_for_timeout(app_hard_wait[short_app_path])
        else:
            # Wait for app to stop updating
            wait_for_idle_app(
                page,
                duration=app_idle_wait["duration"],
                timeout=app_idle_wait["timeout"],
            )

        # Check for py-shiny errors
        error_lines = str(sa.stderr).splitlines()

        # Remove any errors that are allowed
        error_lines = [
            line
            for line in error_lines
            if not any([error_txt in line for error_txt in app_allow_external_errors])
        ]

        # Remove any app specific errors that are allowed
        if short_app_path in app_allow_shiny_errors:
            app_allowable_errors = app_allow_shiny_errors[short_app_path]
        else:
            app_allowable_errors = []

        # If all errors are not allowed, check for unexpected errors
        if app_allowable_errors is not True:
            if isinstance(app_allowable_errors, str):
                app_allowable_errors = [app_allowable_errors]
            app_allowable_errors = (
                # Remove ^INFO lines
                ["INFO:"]
                # Remove any known errors caused by external packages
                + app_allow_external_errors
                # Remove any known errors allowed by the app
                + app_allowable_errors
            )

            # If there is an array of allowable errors, remove them from errors. Ex: `PlotnineWarning`
            error_lines = [
                line
                for line in error_lines
                if len(line.strip()) > 0
                and not any([error_txt in line for error_txt in app_allowable_errors])
            ]
            if len(error_lines) > 0:
                print("\nshort_app_path: " + short_app_path)
                print("\napp_allowable_errors :")
                print("\n".join(app_allowable_errors))
                print("\nError lines remaining:")
                print("\n".join(error_lines))
            assert len(error_lines) == 0

        # Check for JavaScript errors
        if short_app_path in app_allow_js_errors:
            # Remove any errors that are allowed
            console_errors = [
                line
                for line in console_errors
                if not any(
                    [
                        error_txt in line
                        for error_txt in app_allow_js_errors[short_app_path]
                    ]
                )
            ]
        assert len(console_errors) == 0, (
            "In app "
            + ex_app_path
            + " had JavaScript console errors!\n"
            + "* ".join(console_errors)
        )



from __future__ import annotations

import os
import re
import shutil
from pathlib import Path
from typing import Optional


def remove_shinylive_local(
    shinylive_dir: Optional[Path] = None,
    version: Optional[str] = None,
) -> None:
    """Removes local copy of shinylive.

    Parameters
    ----------
    shinylive_dir
        The directory where shinylive is stored. If None, the default directory will
        be used.

    version
        If a version is specified, only that version will be removed.
        If None, all local versions of shinylive will be removed.
    """

    if shinylive_dir is None:
        shinylive_dir = get_default_shinylive_dir()

    target_dir = shinylive_dir
    if version is not None:
        target_dir = target_dir / f"shinylive-{version}"

    if target_dir.exists():
        shutil.rmtree(target_dir)


def get_default_shinylive_dir() -> Path:
    import appdirs

    return (
        Path(
            appdirs.user_cache_dir(  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                "shiny"
            )
        )
        / "shinylive"
    )


def _installed_shinylive_versions(shinylive_dir: Optional[Path] = None) -> list[str]:
    if shinylive_dir is None:
        shinylive_dir = get_default_shinylive_dir()

    if not shinylive_dir.exists():
        return []
    subdirs = next(os.walk(shinylive_dir))[1]
    subdirs = [re.sub("^shinylive-", "", str(s)) for s in subdirs]
    return subdirs


def print_shinylive_local_info() -> None:
    print(
        f"""    Local shinylive dir:
        {get_default_shinylive_dir()}

    Installed versions:
        {", ".join(_installed_shinylive_versions())}"""
    )



from shiny.express import input, render, ui

ui.page_opts(fillable=True)

with ui.card():
    with ui.layout_sidebar():
        with ui.sidebar(id="sidebar_left", open="desktop"):
            "Left sidebar content"

        @render.code
        def state_left():
            return f"input.sidebar_left(): {input.sidebar_left()}"


with ui.card():
    with ui.layout_sidebar():
        with ui.sidebar(id="sidebar_right", position="right", open="desktop"):
            "Right sidebar content"

        @render.code
        def state_right():
            return f"input.sidebar_right(): {input.sidebar_right()}"


with ui.card():
    with ui.layout_sidebar():
        with ui.sidebar(id="sidebar_closed", open="closed"):
            "Closed sidebar content"

        @render.code
        def state_closed():
            return f"input.sidebar_closed(): {input.sidebar_closed()}"


with ui.card():
    with ui.layout_sidebar():
        with ui.sidebar(id="sidebar_always", open="always"):
            "Always sidebar content"

        @render.code
        def state_always():
            return f"input.sidebar_always(): {input.sidebar_always()}"



from shiny.express import input, render, ui
from shiny.types import SilentException

ui.input_text(
    "txt",
    "Enter text to see it displayed below the input",
    width="400px",
)


@render.ui
def txt_out():
    if not input.txt():
        raise SilentException()
    return "Your input: " + input.txt()



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_output_image_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    controller.OutputTextVerbatim(page, "no_output").expect_value("DATA SUMMARY")
    controller.OutputTextVerbatim(page, "no_parens").expect_value("DATA SUMMARY")
    controller.OutputTextVerbatim(page, "to_upper").expect_value("DATA SUMMARY")
    controller.OutputTextVerbatim(page, "to_lower").expect_value("data summary")



from shiny import App, ui

app_ui = ui.page_fluid(
    ui.card(
        ui.card_header("This is the header"),
        ui.card_body(
            ui.tags.h5("This is the title"),
            ui.p("This is the body."),
            ui.p("This is still the body."),
        ),
        ui.card_footer("This is the footer"),
        full_screen=True,
    ),
    ui.card(
        ui.p("These first two elements will be wrapped in `card_body()` together."),
        ui.p("These first two elements will be wrapped in `card_body()` together."),
        ui.card_body(ui.p("A card body.")),
        ui.p("These last two elements will be wrapped in `card_body()` together."),
        ui.p("These last two elements will be wrapped in `card_body()` together."),
    ),
)


app = App(app_ui, server=None)



from shiny import App, Inputs, Outputs, Session, reactive, render, ui

app_ui = ui.page_fluid(
    ui.input_password("password", "Password:"),
    ui.input_action_button("go", "Go"),
    ui.output_text_verbatim("value"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.text
    @reactive.event(input.go)
    def value():
        return input.password()


app = App(app_ui, server)



from shared import filler_text

from shiny.express import input, render, ui

ui.page_opts(
    title="Theme Example",
    theme="https://cdn.jsdelivr.net/npm/bootswatch@5.3.3/dist/sketchy/bootstrap.min.css",
)

with ui.sidebar(title="Parameters"):
    ui.input_numeric("n", "N", min=0, max=100, value=20)

ui.h2("Output")


@render.code
def txt():
    return f"n*2 is {input.n() * 2}"


ui.markdown(filler_text)



from __future__ import annotations

from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_row_selection(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    df = controller.OutputDataFrame(page, "df1")
    add_row = controller.InputActionButton(page, "add_row")
    clear_table = controller.InputActionButton(page, "clear_table")
    selected_rows = controller.OutputTextVerbatim(page, "number_of_selected_rows")

    df.expect_nrow(0)
    selected_rows.expect_value("Selected rows: 0")

    add_row.click()

    df.expect_nrow(1)
    selected_rows.expect_value("Selected rows: 0")

    df.cell_locator(0, 0).click()
    df.select_rows([0])

    df.expect_nrow(1)
    selected_rows.expect_value("Selected rows: 1")

    clear_table.click()
    selected_rows.expect_value("Selected rows: 0")

    bad_error_lines = [line for line in local_app.stderr._lines if "INFO:" not in line]
    assert len(bad_error_lines) == 0, bad_error_lines



from __future__ import annotations

import sys
import tempfile
from pathlib import Path
from typing import Any

import pytest

from shiny import render, ui
from shiny.express import output_args
from shiny.express import ui as xui
from shiny.express._run import run_express


def test_express_ui_is_complete():
    """
    Make sure shiny.express.ui covers everything that shiny.ui does, or explicitly lists
    the item in `_known_missing`.
    These entries are in `_known_missing` in shiny/express/ui/__init__.py
    """

    ui_all = set(ui.__all__)
    xui_all = set(xui.__all__)
    ui_known_missing = set(xui._known_missing["shiny.ui"])
    xui_known_missing = set(xui._known_missing["shiny.express.ui"])

    # Make sure that there's no overlap between shiny.express.ui.__all__ and
    # _known_missing["shiny.ui"]; and same for other combinations. Note that the use of
    # `.intersection() == set()` instead of `disjoint()` is intentional, because if the
    # test fails, the first form provides an error message that shows the difference,
    # while the second form does note.
    assert xui_all.intersection(ui_known_missing) == set()
    assert ui_all.intersection(xui_known_missing) == set()

    # Similar to above, use .difference() instead of .issubset() to get better error
    # messages.
    assert xui_known_missing.difference(xui_all) == set()
    assert ui_known_missing.difference(ui_all) == set()

    # Make sure that everything from shiny.ui is either exported by shiny.express.ui, or
    # explicitly listed in shiny.express.ui._known_missing.
    assert ui_all.union(xui_known_missing) == xui_all.union(ui_known_missing)


def test_render_output_controls():
    @render.text
    def text1():
        return "text"

    assert (
        ui.TagList(text1.tagify()).get_html_string()
        == ui.output_text("text1").get_html_string()
    )

    @output_args(placeholder=False)
    @render.code
    def code1():
        return "text"

    assert (
        ui.TagList(code1.tagify()).get_html_string()
        == ui.output_code("code1", placeholder=False).get_html_string()
    )

    @output_args(width=100)
    @render.code
    def code2():
        return "text"

    with pytest.raises(TypeError, match="width"):
        code2.tagify()


def test_hold():
    old_displayhook = sys.displayhook
    try:
        called = False

        def display_hook_spy(_: object) -> Any:
            nonlocal called
            called = True

        sys.displayhook = display_hook_spy

        with xui.hold():
            sys.displayhook("foo")

        assert not called

        sys.displayhook("baz")
        assert called

        called = False
        with xui.hold() as held:
            sys.displayhook("foo")
        assert not called
        sys.displayhook(held)
        assert called

    finally:
        sys.displayhook = old_displayhook


def test_recall_context_manager():
    # A Shiny Express app that uses a RecallContextManager (ui.card_header()) without
    # `with`. It is used within another RecallContextManager (ui.card()), but that one
    # is used with `with`. This test makes sure that the non-with RecallContextManager
    # will invoke the wrapped function and its result will be passed to the parent.

    card_app_express_text = """\
from shiny.express import ui

with ui.card():
    ui.card_header("Header")
    "Body"
"""

    # The same UI, written in the Shiny Core style.
    card_app_core = ui.page_fixed(
        ui.card(
            ui.card_header("Header"),
            "Body",
        )
    )

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file = Path(temp_dir, "temp.file")
        temp_file.write_text(card_app_express_text)
        res = run_express(temp_file).tagify()

    assert str(res) == str(card_app_core)



from shiny.express import expressify, input, render, ui


@expressify
def my_accordion(**kwargs):
    with ui.accordion(**kwargs):
        for letter in "ABCDE":
            with ui.accordion_panel(f"Section {letter}"):
                f"Some narrative for section {letter}"


ui.markdown("#### Single-select accordion")

my_accordion(multiple=False, id="acc_single")


@render.code
def acc_single_val():
    return "input.acc_single(): " + str(input.acc_single())


ui.br()

ui.markdown("#### Multi-select accordion")

my_accordion(multiple=True, id="acc_multiple")


@render.code
def acc_multiple_val():
    return "input.acc_multiple(): " + str(input.acc_multiple())



import matplotlib.pyplot as plt
import numpy as np

from shiny import ui


def plot_loss_over_time():
    epochs = np.arange(1, 101)
    loss = 1000 / np.sqrt(epochs) + np.random.rand(100) * 25

    fig = plt.figure(figsize=(10, 6))
    plt.plot(epochs, loss)
    plt.xlabel("Epochs")
    plt.ylabel("Loss")
    return fig


def plot_accuracy_over_time():
    epochs = np.arange(1, 101)
    accuracy = np.sqrt(epochs) / 12 + np.random.rand(100) * 0.15
    accuracy = [np.min([np.max(accuracy[:i]), 1]) for i in range(1, 101)]

    fig = plt.figure(figsize=(10, 6))
    plt.plot(epochs, accuracy)
    plt.xlabel("Epochs")
    plt.ylabel("Accuracy")
    return fig


def plot_feature_importance():
    features = ["Product Category", "Price", "Brand", "Rating", "Number of Reviews"]
    importance = np.random.rand(5)

    fig = plt.figure(figsize=(10, 6))
    plt.barh(features, importance)
    plt.xlabel("Importance")
    return fig


card_loss = ui.card(
    ui.card_header("Loss Over Time"),
    ui.output_plot("loss_over_time"),
    full_screen=True,
)

card_acc = ui.card(
    ui.card_header("Accuracy Over Time"),
    ui.output_plot("accuracy_over_time"),
    full_screen=True,
)

card_feat = ui.card(
    ui.card_header("Feature Importance"),
    ui.output_plot("feature_importance"),
    full_screen=True,
)
